# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi
# Based on test_examples.py from CVXOPT by M. Andersen and L. Vandenberghe.

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import math

import numpy as np

import qics


def inp(x, y):
    return sum([np.sum(xi * yi.conj()) for (xi, yi) in zip(x, y)]).real


def vec(x):
    return np.hstack([xk.view(np.float64).ravel() for xk in x])


def rand(dims, types):
    def rand_single(dim, type):
        if type == "r":
            return np.random.randn(dim, 1)
        elif type == "s":
            n = math.isqrt(dim)
            X = np.random.randn(n, n)
            return X + X.T
        elif type == "h":
            n = math.isqrt(dim // 2)
            X = np.random.randn(n, n) + np.random.randn(n, n) * 1j
            return X + X.conj().T

    return [rand_single(dim, type) for (dim, type) in zip(dims, types)]


def pos(dims, types):
    def rand_single(dim, type):
        if type == "r":
            return np.abs(np.random.randn(dim, 1))
        elif type == "s":
            n = math.isqrt(dim)
            X = np.random.randn(n, n)
            return X @ X.T
        elif type == "h":
            n = math.isqrt(dim // 2)
            X = np.random.randn(n, n) + np.random.randn(n, n) * 1j
            return X @ X.conj().T

    return [rand_single(dim, type) for (dim, type) in zip(dims, types)]


def zeros(dims, types):
    def zeros_single(dim, type):
        if type == "r":
            return np.zeros((dim, 1))
        elif type == "s":
            n = math.isqrt(dim)
            return np.zeros((n, n))
        elif type == "h":
            n = math.isqrt(dim // 2)
            return np.zeros((n, n), dtype=np.complex128)

    return [zeros_single(dim, type) for (dim, type) in zip(dims, types)]


def isclose(x, y, tol=1e-3):
    if isinstance(x, np.ndarray):
        return (
            np.linalg.norm(x - y, ord=np.inf) / (1 + np.linalg.norm(x, ord=np.inf))
            <= tol
        )
    else:
        return np.abs(x - y) / (1 + np.abs(x)) <= tol


def _test_cone(K):
    # Initialize at random feasible point
    while True:
        x0 = pos(K.dim, K.type)
        K.set_point(x0, x0)

        if K.get_feas():
            break

    # Get random direction
    H = rand(K.dim, K.type)

    # Compute gradient, Hessian, and TOA at x0
    eps = 1e-8
    f0 = K.get_val()
    g0 = zeros(K.dim, K.type)
    K.grad_ip(g0)
    H0 = zeros(K.dim, K.type)
    K.hess_prod_ip(H0, H)
    T0 = zeros(K.dim, K.type)
    K.third_dir_deriv_axpy(T0, H)

    # Get finite differenced gradient
    f1 = zeros(K.dim, K.type)
    g_fd = zeros(K.dim, K.type)
    for k in range(len(f1)):
        if K.type[k] == "r":
            for i in range(K.dim[k]):
                x1 = [xk.copy() for xk in x0]
                x1[k][i] += eps
                K.set_point(x1, x1)
                K.get_feas()
                f1[k][i] = K.get_val()
        elif K.type[k] == "s" or K.type[k] == "h":
            n = math.isqrt(K.dim[k]) if K.type[k] == "s" else math.isqrt(K.dim[k] // 2)
            for j in range(n):
                for i in range(j + 1):
                    x1 = [xk.copy() for xk in x0]
                    x1[k][i, j] += eps * 0.5
                    x1[k][j, i] += eps * 0.5
                    K.set_point(x1, x1)
                    K.get_feas()
                    f1[k][i, j] = K.get_val()
                    f1[k][j, i] = K.get_val()

                    if K.type[k] == "h":
                        x1 = [xk.copy() for xk in x0]
                        x1[k][i, j] += eps * 0.5j
                        x1[k][j, i] -= eps * 0.5j
                        K.set_point(x1, x1)
                        K.get_feas()
                        f1[k][i, j] += K.get_val() * 1j
                        f1[k][j, i] -= K.get_val() * 1j

        if K.type[k] == "r" or K.type[k] == "s":
            g_fd[k] = (f1[k] - f0) / eps
        elif K.type[k] == "h":
            mask = np.ones_like(g_fd[k])
            mask[np.triu_indices(mask.shape[0], k=1)] += 1j
            mask[np.tril_indices(mask.shape[0], k=-1)] -= 1j
            g_fd[k] = (f1[k] - mask * f0) / eps

    # Compute gradient, Hessian, and TOA at x0
    x1 = [xk + eps * Hk for (xk, Hk) in zip(x0, H)]
    K.set_point(x1, x1)
    K.get_feas()
    g1 = zeros(K.dim, K.type)
    K.grad_ip(g1)
    H1 = zeros(K.dim, K.type)
    K.hess_prod_ip(H1, H)
    T1 = zeros(K.dim, K.type)
    K.third_dir_deriv_axpy(T1, H)

    # Gradient checks
    assert isclose(
        (vec(g0) + vec(g1)) / 2, vec(g_fd)
    ), f"{K} gradient oracle does not match finite differencing approximation"
    assert isclose(
        inp(g0, x0), -K.nu
    ), f"{K} gradient oracle does not satisfy <g(x), x> = -nu identity"

    # Hessian checks
    work = zeros(K.dim, K.type)
    K.hess_prod_ip(work, x0)
    assert isclose(
        vec(work), -vec(g0)
    ), f"{K} Hessian oracle does not satisfy H(x)x = -g(x) identity"
    assert isclose(
        inp(K.hess_prod_ip(work, x0), x0), K.nu
    ), f"{K} Hessian oracle does not satisfy <H(x)x, x> = nu identity"
    assert isclose(
        (vec(H0) + vec(H1)) / 2, (vec(g1) - vec(g0)) / eps
    ), f"{K} Hessian oracle does not match finite differencing approximation"

    # Inverse Hessian checks
    K.invhess_prod_ip(work, H1)
    assert isclose(
        vec(H), vec(work)
    ), f"{K} inverse Hessian oracle does not satisfy H^-1(H(h)) = h identity"
    assert isclose(
        inp(K.invhess_prod_ip(work, g0), g0), K.nu
    ), f"{K} inverse Hessian oracle does not satisfy <H^-1(g(x)), g(x)> = nu identity"

    # Third order derivative checks
    assert isclose(
        (vec(T0) + vec(T1)) / 2, (vec(H1) - vec(H0)) / eps
    ), f"{K} third order derivative oracle does not match finite differencing approximation"


def test_cone_oracles():
    np.random.seed(1)

    _test_cone(qics.cones.NonNegOrthant(3))
    _test_cone(qics.cones.SecondOrder(3))
    _test_cone(qics.cones.PosSemidefinite(3))
    _test_cone(qics.cones.PosSemidefinite(3, True))

    _test_cone(qics.cones.ClassEntr(3))
    _test_cone(qics.cones.ClassRelEntr(3))

    _test_cone(qics.cones.QuantEntr(3))
    _test_cone(qics.cones.QuantEntr(3, True))

    _test_cone(qics.cones.QuantRelEntr(3))
    _test_cone(qics.cones.QuantRelEntr(3, True))

    _test_cone(qics.cones.QuantCondEntr((2, 2, 2, 2), (1, 3)))
    _test_cone(qics.cones.QuantCondEntr((2, 2, 2, 2), (1, 3), True))

    K_list = [np.random.randn(4, 2), np.random.randn(4, 2)]
    _test_cone(qics.cones.QuantKeyDist(K_list, 2))
    _test_cone(qics.cones.QuantKeyDist(K_list, 2, True))

    _test_cone(qics.cones.OpPerspecTr(3, "log"))
    _test_cone(qics.cones.OpPerspecTr(3, "log", True))
    _test_cone(qics.cones.OpPerspecTr(3, 0.3))
    _test_cone(qics.cones.OpPerspecTr(3, 0.3, True))

    _test_cone(qics.cones.OpPerspecEpi(3, "log"))
    _test_cone(qics.cones.OpPerspecEpi(3, "log", True))
    _test_cone(qics.cones.OpPerspecEpi(3, 0.3))
    _test_cone(qics.cones.OpPerspecEpi(3, 0.3, True))



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numba as nb
import numpy as np

from qics._utils.linalg import congr_multi


@nb.njit(cache=True)
def D1_f(D, f_D, df_D):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)

    n = D.size
    D1 = np.empty((n, n))

    for j in range(n):
        for i in range(j):
            d_ij = D[i] - D[j]
            if abs(d_ij) < rteps:
                D1[i, j] = 0.5 * (df_D[i] + df_D[j])
            else:
                D1[i, j] = (f_D[i] - f_D[j]) / d_ij
            D1[j, i] = D1[i, j]

        D1[j, j] = df_D[j]

    return D1


@nb.njit(cache=True)
def D1_log(D, log_D):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)

    n = D.size
    D1 = np.empty((n, n))

    for j in range(n):
        for i in range(j):
            d_ij = D[i] - D[j]
            if abs(d_ij) < rteps:
                D1[i, j] = 2.0 / (D[i] + D[j])
            elif (D[i] < D[j] / 2) or (D[j] < D[i] / 2):
                D1[i, j] = (log_D[i] - log_D[j]) / d_ij
            else:
                z = d_ij / (D[i] + D[j])
                D1[i, j] = 2.0 * np.arctanh(z) / d_ij
            D1[j, i] = D1[i, j]

        D1[j, j] = np.reciprocal(D[j])

    return D1


@nb.njit(cache=True)
def D2_f(D, D1, d2f_D):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)

    n = D.size
    D2 = np.zeros((n, n, n))

    for k in range(n):
        for j in range(k + 1):
            for i in range(j + 1):
                d_jk = D[j] - D[k]
                if abs(d_jk) < rteps:
                    d_ij = D[i] - D[j]
                    if abs(d_ij) < rteps:
                        t = (d2f_D[i] + d2f_D[j] + d2f_D[k]) / 6
                    else:
                        t = (D1[i, j] - D1[j, k]) / d_ij
                else:
                    t = (D1[i, j] - D1[i, k]) / d_jk

                D2[i, j, k] = t
                D2[i, k, j] = t
                D2[j, i, k] = t
                D2[j, k, i] = t
                D2[k, i, j] = t
                D2[k, j, i] = t

    return D2


@nb.njit(cache=True)
def D2_log(D, D1):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)

    n = D.size
    D2 = np.zeros((n, n, n))

    for k in range(n):
        for j in range(k + 1):
            for i in range(j + 1):
                d_jk = D[j] - D[k]
                if abs(d_jk) < rteps:
                    d_ij = D[i] - D[j]
                    if abs(d_ij) < rteps:
                        t = ((3 / (D[i] + D[j] + D[k])) ** 2) / -2
                    else:
                        t = (D1[i, j] - D1[j, k]) / d_ij
                else:
                    t = (D1[i, j] - D1[i, k]) / d_jk

                D2[i, j, k] = t
                D2[i, k, j] = t
                D2[j, i, k] = t
                D2[j, k, i] = t
                D2[k, i, j] = t
                D2[k, j, i] = t

    return D2


@nb.njit(cache=True)
def D3_f_ij(i, j, D, D2, d3f_D):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)
    n = D.size
    D_i = D[i]
    D_j = D[j]

    D3_ij = np.zeros((n, n))

    for a in range(n):
        for b in range(a + 1):
            D_b = D[b]
            D_a = D[a]
            D_ij = D_i - D_j
            D_ib = D_i - D_b
            D_ia = D_i - D_a
            B_ib = abs(D_ib) < rteps
            B_ia = abs(D_ia) < rteps

            if (abs(D_ij) < rteps) and B_ib and B_ia:
                t = d3f_D[i] / 6
            elif B_ib and B_ia:
                t = (D2[i, i, i] - D2[i, i, j]) / D_ij
            elif B_ia:
                t = (D2[i, i, j] - D2[i, j, b]) / D_ib
            else:
                t = (D2[i, j, b] - D2[j, b, a]) / D_ia

            D3_ij[b, a] = t
            D3_ij[a, b] = t

    return D3_ij


@nb.njit(cache=True)
def D3_log_ij(i, j, D2, D, f):
    eps = np.finfo(np.float64).eps
    rteps = np.sqrt(eps)
    n = D.size
    D_i = D[i]
    D_j = D[j]

    D3_ij = np.zeros((n, n))

    for a in range(n):
        for b in range(a + 1):
            D_b = D[b]
            D_a = D[a]
            D_ij = D_i - D_j
            D_ib = D_i - D_b
            D_ia = D_i - D_a
            B_ib = abs(D_ib) < rteps
            B_ia = abs(D_ia) < rteps

            if (abs(D_ij) < rteps) and B_ib and B_ia:
                t = D_i**-3 / 3 if (f == "log") else -(D_i**-2) / 6
            elif B_ib and B_ia:
                t = (D2[i, i, i] - D2[i, i, j]) / D_ij
            elif B_ia:
                t = (D2[i, i, j] - D2[i, j, b]) / D_ib
            else:
                t = (D2[i, j, b] - D2[j, b, a]) / D_ia

            D3_ij[b, a] = t
            D3_ij[a, b] = t

    return D3_ij


def scnd_frechet(D2, UHU, UXU=None, U=None):
    n = D2.shape[0]

    D2_UXU = (D2 * UXU) if (UXU is not None) else (D2)
    out = UHU.reshape((n, 1, n)) @ D2_UXU
    out = out.reshape((n, n))
    out = out + out.conj().T
    out = (U @ out @ U.conj().T) if (U is not None) else (out)

    return out


def scnd_frechet_multi(
    out, D2, UHU, UXU=None, U=None, work1=None, work2=None, work3=None
):
    D2_UXU = (D2 * UXU) if (UXU is not None) else (D2)
    np.matmul(D2_UXU, UHU.conj().transpose(1, 2, 0), out=work3)

    if U is not None:
        np.add(work3.transpose(2, 1, 0), work3.conj().transpose(2, 0, 1), out=work1)
        congr_multi(out, U, work1, work2)
    else:
        np.add(work3.transpose(2, 1, 0), work3.conj().transpose(2, 0, 1), out=out)

    return out


@nb.njit(cache=True, parallel=True)
def thrd_frechet(D, D2, d3f_D, U, H1, H2, H3=None):
    n = D.size
    out = np.zeros_like(H1)

    # If H3 is None, then assume H2=H3
    if H3 is None:
        for i in nb.prange(n):
            for j in range(i + 1):
                D3_ij = D3_f_ij(i, j, D, D2, d3f_D)

                for b in range(n):
                    for a in range(n):
                        temp = H1[i, b] * H2[b, a] * H2[a, j]
                        temp += H2[i, b] * (H1[b, a] * H2[a, j] + H2[b, a] * H1[a, j])
                        out[i, j] = out[i, j] + D3_ij[b, a] * temp

                out[j, i] = np.conj(out[i, j])

        out *= 2

    else:
        for i in nb.prange(n):
            for j in range(i + 1):
                D3_ij = D3_f_ij(i, j, D, D2, d3f_D)

                for b in range(n):
                    for a in range(n):
                        work = H2[b, a] * H3[a, j]
                        work += H3[b, a] * H2[a, j]
                        temp = H1[i, b] * work

                        work = H1[b, a] * H3[a, j]
                        work += H3[b, a] * H1[a, j]
                        temp += H2[i, b] * work

                        work = H1[b, a] * H2[a, j]
                        work += H2[b, a] * H1[a, j]
                        temp += H3[i, b] * work
                        out[i, j] = out[i, j] + D3_ij[b, a] * temp

                out[j, i] = np.conj(out[i, j])

    return U @ out @ U.conj().T


def get_S_matrix(D2_UXU, rt2, iscomplex=False):
    if iscomplex:
        return get_S_matrix_hermitian(D2_UXU, rt2)
    else:
        return get_S_matrix_symmetric(D2_UXU, rt2)


@nb.njit(cache=True)
def get_S_matrix_symmetric(D2_UXU, rt2):
    n = D2_UXU.shape[0]
    vn = n * (n + 1) // 2
    S = np.zeros((vn, vn))
    col = 0

    for j in range(n):
        for i in range(j):
            # Column corresponding to unit vector (Hij + Hji) / sqrt(2)
            # Increment rows
            for k in range(j):
                row = k + (j * (j + 1)) // 2
                S[row, col] = D2_UXU[j, k, i]
            row = j + (j * (j + 1)) // 2
            S[row, col] = D2_UXU[j, j, i] * rt2
            for k in range(j + 1, n):
                row = j + (k * (k + 1)) // 2
                S[row, col] = D2_UXU[j, k, i]

            # Increment columns
            for k in range(i):
                row = k + (i * (i + 1)) // 2
                S[row, col] += D2_UXU[i, k, j]
            row = i + (i * (i + 1)) // 2
            S[row, col] = D2_UXU[i, j, i] * rt2
            for k in range(i + 1, n):
                row = i + (k * (k + 1)) // 2
                S[row, col] += D2_UXU[i, k, j]

            col += 1

        # Column corresponding to unit vector Hjj
        for k in range(j):
            row = k + (j * (j + 1)) // 2
            S[row, col] = D2_UXU[j, j, k] * rt2
        row = j + (j * (j + 1)) // 2
        S[row, col] = 2 * D2_UXU[j, j, j]
        for k in range(j + 1, n):
            row = j + (k * (k + 1)) // 2
            S[row, col] = D2_UXU[j, j, k] * rt2

        col += 1

    return S


@nb.njit(cache=True)
def get_S_matrix_hermitian(D2_UXU, rt2):
    n = D2_UXU.shape[0]
    vn = n * n
    S = np.zeros((vn, vn))
    col = 0

    for j in range(n):
        for i in range(j):
            # Column corresponding to unit vector (Hij + Hji) / sqrt(2)
            # Increment rows
            for k in range(j):
                row = 2 * k + j * j
                S[row, col] = D2_UXU[j, k, i].real
                S[row + 1, col] = D2_UXU[j, k, i].imag
            row = 2 * j + j * j
            S[row, col] = D2_UXU[j, j, i].real * rt2
            for k in range(j + 1, n):
                row = 2 * j + k * k
                S[row, col] = D2_UXU[j, k, i].real
                S[row + 1, col] = -D2_UXU[j, k, i].imag

            # Increment columns
            for k in range(i):
                row = 2 * k + i * i
                S[row, col] += D2_UXU[i, k, j].real
                S[row + 1, col] += D2_UXU[i, k, j].imag
            row = 2 * i + i * i
            S[row, col] = D2_UXU[i, j, i].real * rt2
            for k in range(i + 1, n):
                row = 2 * i + k * k
                S[row, col] += D2_UXU[i, k, j].real
                S[row + 1, col] += -D2_UXU[i, k, j].imag

            col += 1

            # Increment rows
            for k in range(j):
                row = 2 * k + j * j
                S[row, col] = -D2_UXU[j, k, i].imag
                S[row + 1, col] = D2_UXU[j, k, i].real
            row = 2 * j + j * j
            S[row, col] = -D2_UXU[j, j, i].imag * rt2
            for k in range(j + 1, n):
                row = 2 * j + k * k
                S[row, col] = -D2_UXU[j, k, i].imag
                S[row + 1, col] = -D2_UXU[j, k, i].real

            # Increment columns
            for k in range(i):
                row = 2 * k + i * i
                S[row, col] += D2_UXU[i, k, j].imag
                S[row + 1, col] += -D2_UXU[i, k, j].real
            row = 2 * i + i * i
            S[row, col] = -D2_UXU[i, j, i].imag * rt2
            for k in range(i + 1, n):
                row = 2 * i + k * k
                S[row, col] += D2_UXU[i, k, j].imag
                S[row + 1, col] += D2_UXU[i, k, j].real

            col += 1

        # Column corresponding to unit vector Hjj
        for k in range(j):
            row = 2 * k + j * j
            S[row, col] = D2_UXU[j, j, k].real * rt2
            S[row + 1, col] = -D2_UXU[j, j, k].imag * rt2
        row = 2 * j + j * j
        S[row, col] = 2 * D2_UXU[j, j, j].real
        for k in range(j + 1, n):
            row = 2 * j + k * k
            S[row, col] = D2_UXU[j, j, k].real * rt2
            S[row + 1, col] = D2_UXU[j, j, k].imag * rt2

        col += 1

    return S



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.gradient import (
    D1_f,
    D1_log,
    D2_f,
    scnd_frechet,
    scnd_frechet_multi,
    thrd_frechet,
)
from qics._utils.linalg import (
    cho_fact,
    cho_solve,
    congr_multi,
    dense_dot_x,
    inp,
    x_dot_dense,
)
from qics.cones.base import Cone, get_perspective_derivatives
from qics.vectorize import get_full_to_compact_op, vec_to_mat


class OpPerspecTr(Cone):
    r"""A class representing a trace operator perspective epigraph cone

    .. math::

        \mathcal{OPT}_{n, g} = \text{cl}\{ (t, X, Y) \in \mathbb{R} \times
        \mathbb{H}^n_{++} \times \mathbb{H}^n_{++} :
        t \geq \text{tr}[P_g(X, Y)] \},

    for an operator concave function
    :math:`g:(0, \infty)\rightarrow\mathbb{R}`, where

    .. math::

        P_g(X, Y) = X^{1/2} g(X^{-1/2} Y X^{-1/2}) X^{1/2},

    is the operator perspective of :math:`g`.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the matrices :math:`X` and :math:`Y`.
    func : :obj:`string` or :obj:`float`
        Choice for the function :math:`g`. Can be defined in the following
        ways.

        - :math:`g(x) = -\log(x)` if ``func="log"``
        - :math:`g(x) = -x^p` if ``func=p`` is a :obj:`float` where
          :math:`p\in(0, 1)`
        - :math:`g(x) = x^p` if ``func=p`` is a :obj:`float` where
          :math:`p\in[-1, 0)\cup(1, 2)`

    iscomplex : :obj:`bool`
        Whether the matrices :math:`X` and :math:`Y` are defined over
        :math:`\mathbb{H}^n` (``True``), or restricted to
        :math:`\mathbb{S}^n` (``False``). The default is ``False``.

    See also
    --------
    OpPerspecEpi : Operator perspective epigraph

    Notes
    -----
    We do not support operator perspectives for ``p=0``, ``p=1``, and
    ``p=2`` as these functions are more efficiently modelled using
    just the positive semidefinite cone.

    - When :math:`g(x)=x^0`, :math:`P_g(X, Y)=X`.
    - When :math:`g(x)=x^1`, :math:`P_g(X, Y)=Y`.
    - When :math:`g(x)=x^2`, :math:`P_g(X, Y)=YX^{-1}Y`, which can be
      modelled using the Schur complement lemma, i.e., if :math:`X\succ 0`,
      then

      .. math::

          \begin{bmatrix} X & Y \\ Y & T \end{bmatrix} \succeq 0 \qquad
          \Longleftrightarrow \qquad T \succeq YX^{-1}Y.
    """

    def __init__(self, n, func, iscomplex=False):
        self.n = n
        self.func = func
        self.iscomplex = iscomplex

        self.nu = 1 + 2 * self.n  # Barrier parameter

        if iscomplex:
            self.vn = n * n
            self.dim = [1, 2 * n * n, 2 * n * n]
            self.type = ["r", "h", "h"]
            self.dtype = np.complex128
        else:
            self.vn = n * (n + 1) // 2
            self.dim = [1, n * n, n * n]
            self.type = ["r", "s", "s"]
            self.dtype = np.float64

        self.idx_X = slice(1, 1 + self.dim[1])
        self.idx_Y = slice(1 + self.dim[1], sum(self.dim))

        # Get function handles for g(x), h(x)=x*g(1/x), x*g(x), and x*h(x)
        # and their first, second and third derivatives
        perspective_derivatives = get_perspective_derivatives(func)
        self.g, self.dg, self.d2g, self.d3g = perspective_derivatives["g"]
        self.h, self.dh, self.d2h, self.d3h = perspective_derivatives["h"]
        self.xg, self.dxg, self.d2xg, self.d3xg = perspective_derivatives["xg"]
        self.xh, self.dxh, self.d2xh, self.d3xh = perspective_derivatives["xh"]

        # Get sparse operator to convert from full to compact vectorizations
        self.F2C_op = get_full_to_compact_op(n, iscomplex)

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.hess_aux_updated = False
        self.invhess_aux_updated = False
        self.invhess_aux_aux_updated = False
        self.dder3_aux_updated = False
        self.congr_aux_updated = False

        return

    def get_iscomplex(self):
        return self.iscomplex

    def get_init_point(self, out):
        (t0, x0, y0) = self.get_central_ray()

        point = [
            np.array([[t0]]),
            np.eye(self.n, dtype=self.dtype) * x0,
            np.eye(self.n, dtype=self.dtype) * y0,
        ]

        self.set_point(point, point)

        out[0][:] = point[0]
        out[1][:] = point[1]
        out[2][:] = point[2]

        return out

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        (self.t, self.X, self.Y) = self.primal

        # Check that X and Y are positive definite
        self.Dx, self.Ux = np.linalg.eigh(self.X)
        self.Dy, self.Uy = np.linalg.eigh(self.Y)

        if any(self.Dx <= 0) or any(self.Dy <= 0):
            self.feas = False
            return self.feas

        # Construct (X^-1/2 Y X^-1/2) and (Y^-1/2 X Y^-1/2)
        # and double check they are also PSD (in case of numerical errors)
        rt2_Dx = np.sqrt(self.Dx)
        rt4_X = self.Ux * np.sqrt(rt2_Dx)
        irt4_X = self.Ux / np.sqrt(rt2_Dx)
        self.rt2_X = rt4_X @ rt4_X.conj().T
        self.irt2_X = irt4_X @ irt4_X.conj().T

        rt2_Dy = np.sqrt(self.Dy)
        rt4_Y = self.Uy * np.sqrt(rt2_Dy)
        irt4_Y = self.Uy / np.sqrt(rt2_Dy)
        self.rt2_Y = rt4_Y @ rt4_Y.conj().T
        self.irt2_Y = irt4_Y @ irt4_Y.conj().T

        XYX = self.irt2_X @ self.Y @ self.irt2_X
        YXY = self.irt2_Y @ self.X @ self.irt2_Y

        self.Dxyx, self.Uxyx = np.linalg.eigh(XYX)
        self.Dyxy, self.Uyxy = np.linalg.eigh(YXY)

        if any(self.Dxyx <= 0) or any(self.Dyxy <= 0):
            self.feas = False
            return self.feas

        # Check that t > tr[Pg(X, Y)]
        self.g_Dxyx = self.g(self.Dxyx)
        self.h_Dyxy = self.h(self.Dyxy)
        g_XYX = (self.Uxyx * self.g_Dxyx) @ self.Uxyx.conj().T
        self.z = self.t[0, 0] - inp(self.X, g_XYX)

        self.feas = self.z > 0
        return self.feas

    def get_val(self):
        assert self.feas_updated
        log_z = np.log(self.z)
        return -log_z - np.sum(np.log(self.Dx)) - np.sum(np.log(self.Dy))

    def update_grad(self):
        assert self.feas_updated
        assert not self.grad_updated

        # Compute X^-1 and Y^-1
        inv_Dx = np.reciprocal(self.Dx)
        inv_X_rt2 = self.Ux * np.sqrt(inv_Dx)
        self.inv_X = inv_X_rt2 @ inv_X_rt2.conj().T

        inv_Dy = np.reciprocal(self.Dy)
        inv_Y_rt2 = self.Uy * np.sqrt(inv_Dy)
        self.inv_Y = inv_Y_rt2 @ inv_Y_rt2.conj().T

        # Precompute useful expressions
        self.UyxyYUyxy = UyxyYUyxy = self.Uyxy.conj().T @ self.Y @ self.Uyxy
        self.UxyxXUxyx = UxyxXUxyx = self.Uxyx.conj().T @ self.X @ self.Uxyx

        self.irt2Y_Uyxy = irt2Y_Uyxy = self.irt2_Y @ self.Uyxy
        self.irt2X_Uxyx = irt2X_Uxyx = self.irt2_X @ self.Uxyx

        self.rt2Y_Uyxy = self.rt2_Y @ self.Uyxy
        self.rt2X_Uxyx = self.rt2_X @ self.Uxyx

        # Compute derivatives of trace operator perspective
        self.D1yxy_h = D1_f(self.Dyxy, self.h_Dyxy, self.dh(self.Dyxy))
        if self.func == "log":
            self.D1xyx_g = -D1_log(self.Dxyx, -self.g_Dxyx)
        else:
            self.D1xyx_g = D1_f(self.Dxyx, self.g_Dxyx, self.dg(self.Dxyx))
        # D_X trPg(X, Y) = Y^-½ Dh(Y^-½ X Y^-½)[Y] Y^-½
        work = irt2Y_Uyxy @ (self.D1yxy_h * UyxyYUyxy) @ irt2Y_Uyxy.conj().T
        self.DPhiX = (work + work.conj().T) * 0.5
        # D_Y trPg(X, Y) = X^-½ Dg(X^-½ Y X^-½)[X] X^-½
        work = irt2X_Uxyx @ (self.D1xyx_g * UxyxXUxyx) @ irt2X_Uxyx.conj().T
        self.DPhiY = (work + work.conj().T) * 0.5

        # Precompute compact vectorizations of derivatives
        DPhiX_cvec = self.DPhiX.view(np.float64).reshape(-1, 1)
        DPhiX_cvec = self.F2C_op @ DPhiX_cvec

        DPhiY_cvec = self.DPhiY.view(np.float64).reshape(-1, 1)
        DPhiY_cvec = self.F2C_op @ DPhiY_cvec

        self.DPhi_cvec = np.vstack((DPhiX_cvec, DPhiY_cvec))

        # Compute gradient of barrier function
        self.zi = np.reciprocal(self.z)

        self.grad = [
            -self.zi,
            self.zi * self.DPhiX - self.inv_X,
            self.zi * self.DPhiY - self.inv_Y,
        ]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()

        (Ht, Hx, Hy) = H

        D2yxy_h, D2yxy_xh = self.D2yxy_h, self.D2yxy_xh
        D2xyx_g, D2xyx_xg = self.D2xyx_g, self.D2xyx_xg
        UyxyYUyxy, UxyxXUxyx = self.UyxyYUyxy, self.UxyxXUxyx
        rt2Y_Uyxy, irt2Y_Uyxy = self.rt2Y_Uyxy, self.irt2Y_Uyxy
        rt2X_Uxyx, irt2X_Uxyx = self.rt2X_Uxyx, self.irt2X_Uxyx

        UyxyYHxYUyxy = self.irt2Y_Uyxy.conj().T @ Hx @ self.irt2Y_Uyxy
        UxyxXHyXUxyx = self.irt2X_Uxyx.conj().T @ Hy @ self.irt2X_Uxyx

        # Hessian product of trace operator perspective
        # D2_XX trPg(X, Y)[Hx] = Y^-½ D2h(Y^-½ X Y^½)[Y, Y^-½ Hx Y^-½] Y^-½
        D2PhiXXH = scnd_frechet(D2yxy_h, UyxyYUyxy, UyxyYHxYUyxy, U=irt2Y_Uyxy)
        # D2_XY trPg(X, Y)[Hy]
        #     = -X^-½ D2xg(X^-½ Y X^-½)[X, X^-½ Hy X^-½] X^-½
        #       + X^½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^-½
        #       + X^-½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^½
        work = self.D1xyx_g * UxyxXHyXUxyx
        D2PhiXYH = irt2X_Uxyx @ work @ rt2X_Uxyx.conj().T
        D2PhiXYH += D2PhiXYH.conj().T
        D2PhiXYH -= scnd_frechet(D2xyx_xg, UxyxXUxyx, UxyxXHyXUxyx, U=irt2X_Uxyx)
        # D2_YX trPg(X, Y)[Hx]
        #     = -Y^-½ D2xh(Y^-½ X Y^-½)[Y, Y^-½ Hx Y^-½] Y^-½
        #       + Y^½ Dh(Y^-½ X Y-^½)[Y^-½ Hx Y^-½] Y^-½
        #       + Y^-½ Dh(Y^-½ X Y-^½)[Y^-½ Hx Y^-½] Y^½
        work = self.D1yxy_h * UyxyYHxYUyxy
        D2PhiYXH = irt2Y_Uyxy @ work @ rt2Y_Uyxy.conj().T
        D2PhiYXH += D2PhiYXH.conj().T
        D2PhiYXH -= scnd_frechet(D2yxy_xh, UyxyYUyxy, UyxyYHxYUyxy, U=irt2Y_Uyxy)
        # D2_YY trPg(X, Y)[Hy] = X^-½ D2g(X^-½ Y X^½)[X, X^-½ Hy X^-½] X^-½
        D2PhiYYH = scnd_frechet(D2xyx_g, UxyxXUxyx, UxyxXHyXUxyx, U=irt2X_Uxyx)

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X, Y)[Ht, Hx, Hy]
        #         = (Ht - D_X S(X||Y)[Hx] - D_Y S(X||Y)[Hy]) / z^2
        out_t = Ht - inp(self.DPhiX, Hx) - inp(self.DPhiY, Hy)
        out_t *= self.zi2
        out[0][:] = out_t

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # D2_X F(t, X, Y)[Ht, Hx, Hy]
        #         = -D2_t F(t, X, Y)[Ht, Hx, Hy] * D_X trPg(X, Y)
        #           + (D2_XX trPg(X, Y)[Hx] + D2_XY trPg(X, Y)[Hy]) / z
        #           + X^-1 Hx X^-1
        out_X = -out_t * self.DPhiX
        out_X += self.zi * (D2PhiXYH + D2PhiXXH)
        out_X += self.inv_X @ Hx @ self.inv_X
        out_X = (out_X + out_X.conj().T) * 0.5
        out[1][:] = out_X

        # ==================================================================
        # Hessian products with respect to Y
        # ==================================================================
        # Hessian product of barrier function
        # D2_Y F(t, X, Y)[Ht, Hx, Hy]
        #         = -D2_t F(t, X, Y)[Ht, Hx, Hy] * D_Y S(X||Y)
        #           + (D2_YX S(X||Y)[Hx] + D2_YY S(X||Y)[Hy]) / z
        #           + Y^-1 Hy Y^-1
        out_Y = -out_t * self.DPhiY
        out_Y += self.zi * (D2PhiYXH + D2PhiYYH)
        out_Y += self.inv_Y @ Hy @ self.inv_Y
        out_Y = (out_Y + out_Y.conj().T) * 0.5
        out[2][:] = out_Y

        return out

    def hess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        p = A.shape[0]
        lhs = np.empty((p, sum(self.dim)))

        D2yxy_h, D2yxy_xh = self.D2yxy_h, self.D2yxy_xh
        D2xyx_g, D2xyx_xg = self.D2xyx_g, self.D2xyx_xg
        UyxyYUyxy, UxyxXUxyx = self.UyxyYUyxy, self.UxyxXUxyx
        rt2Y_Uyxy, irt2Y_Uyxy = self.rt2Y_Uyxy, self.irt2Y_Uyxy
        rt2X_Uxyx, irt2X_Uxyx = self.rt2X_Uxyx, self.irt2X_Uxyx

        work0, work1 = self.work0, self.work1
        work2, work3 = self.work2, self.work3
        work4, work5, work6 = self.work4, self.work5, self.work6

        # ======================================================================
        # Hessian products with respect to t
        # ======================================================================
        # D2_t F(t, X, Y)[Ht, Hx, Hy]
        #         = (Ht - D_X S(X||Y)[Hx] - D_Y S(X||Y)[Hy]) / z^2
        DPhiX_vec = self.DPhiX.view(np.float64).reshape((-1, 1))
        DPhiY_vec = self.DPhiY.view(np.float64).reshape((-1, 1))

        out_t = self.At - (self.Ax_vec @ DPhiX_vec).ravel()
        out_t -= (self.Ay_vec @ DPhiY_vec).ravel()
        out_t *= self.zi2

        lhs[:, 0] = out_t

        # ======================================================================
        # Hessian products with respect to X
        # ======================================================================
        # Hessian products of trace operator perspective
        # D2_XX trPg(X, Y)[Hx] = Y^-½ D2h(Y^-½ X Y^½)[Y, Y^-½ Hx Y^-½] Y^-½
        congr_multi(work0, irt2Y_Uyxy.conj().T, self.Ax, work=work3)
        scnd_frechet_multi(work5, D2yxy_h, work0, UyxyYUyxy, U=irt2Y_Uyxy,
                           work1=work3, work2=work4, work3=work6)  # fmt: skip
        # D2_XY trPg(X, Y)[Hy]
        #     = -X^-½ D2xg(X^-½ Y X^-½)[X, X^-½ Hy X^-½] X^-½
        #       + X^½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^-½
        #       + X^-½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^½
        # Second and third terms, i.e., X^½ [ ... ] X^-½ + X^-½ [ ... ] X^½
        congr_multi(work1, irt2X_Uxyx.conj().T, self.Ay, work=work3)
        np.multiply(work1, self.D1xyx_g, out=work2)
        congr_multi(work3, irt2X_Uxyx, work2, work=work4, B=rt2X_Uxyx)
        np.add(work3, work3.conj().transpose(0, 2, 1), out=work2)
        work5 += work2
        # First term, i.e., -X^-½ D2xg(X^-½ Y X^-½)[X, X^-½ Hy X^-½] X^-½
        scnd_frechet_multi(work2, D2xyx_xg, work1, UxyxXUxyx, U=irt2X_Uxyx,
                           work1=work3, work2=work4, work3=work6)  # fmt: skip
        work5 -= work2

        # Hessian product of barrier function
        # D2_X F(t, X, Y)[Ht, Hx, Hy]
        #         = -D2_t F(t, X, Y)[Ht, Hx, Hy] * D_X trPg(X, Y)
        #           + (D2_XX trPg(X, Y)[Hx] + D2_XY trPg(X, Y)[Hy]) / z
        #           + X^-1 Hx X^-1
        work5 *= self.zi
        np.outer(out_t, self.DPhiX, out=work2.reshape((p, -1)))
        work5 -= work2
        congr_multi(work2, self.inv_X, self.Ax, work=work3)
        work5 += work2

        lhs[:, self.idx_X] = work5.reshape((p, -1)).view(np.float64)

        # ==================================================================
        # Hessian products with respect to Y
        # ==================================================================
        # Hessian products of trace operator perspective
        # D2_YY trPg(X, Y)[Hy] = X^-½ D2g(X^-½ Y X^½)[X, X^-½ Hy X^-½] X^-½
        congr_multi(work1, irt2X_Uxyx.conj().T, self.Ay, work=work3)
        scnd_frechet_multi(work5, D2xyx_g, work1, UxyxXUxyx, U=irt2X_Uxyx, 
                           work1=work3, work2=work4, work3=work6)  # fmt: skip

        # D2_YX trPg(X, Y)[Hx]
        #     = -Y^-½ D2xh(Y^-½ X Y^-½)[Y, Y^-½ Hx Y^-½] Y^-½
        #       + Y^½ Dh(Y^-½ X Y-^½)[Y^-½ Hx Y^-½] Y^-½
        #       + Y^-½ Dh(Y^-½ X Y-^½)[Y^-½ Hx Y^-½] Y^½
        # Second and third terms, i.e., Y^½ [ ... ] Y^-½ + Y^-½ [ ... ] Y^½
        congr_multi(work0, irt2Y_Uyxy.conj().T, self.Ax, work=work3)
        np.multiply(work0, self.D1yxy_h, out=work2)
        congr_multi(work3, irt2Y_Uyxy, work2, work=work4, B=rt2Y_Uyxy)
        np.add(work3, work3.conj().transpose(0, 2, 1), out=work2)
        work5 += work2
        # First term, i.e., -Y^-½ D2xh(Y^-½ X Y^-½)[Y, Y^-½ Hx Y^-½] Y^-½
        scnd_frechet_multi(work2, D2yxy_xh, work0, UyxyYUyxy, U=irt2Y_Uyxy,
                           work1=work3, work2=work4, work3=work6)  # fmt: skip
        work5 -= work2

        # Hessian product of barrier function
        # D2_Y F(t, X, Y)[Ht, Hx, Hy]
        #         = -D2_t F(t, X, Y)[Ht, Hx, Hy] * D_Y S(X||Y)
        #           + (D2_YX S(X||Y)[Hx] + D2_YY S(X||Y)[Hy]) / z
        #           + Y^-1 Hy Y^-1
        work5 *= self.zi
        np.outer(out_t, self.DPhiY, out=work2.reshape((p, -1)))
        work5 -= work2
        congr_multi(work2, self.inv_Y, self.Ay, work=work3)
        work5 += work2

        lhs[:, self.idx_Y] = work5.reshape((p, -1)).view(np.float64)

        # Multiply A (H A')
        return dense_dot_x(lhs, A.T)

    def invhess_prod_ip(self, out, H):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()

        (Ht, Hx, Hy) = H

        # Compute Wx and get compact vectorization
        Wx = Hx + Ht * self.DPhiX
        Wx_vec = Wx.view(np.float64).reshape(-1, 1)
        Wx_cvec = self.F2C_op @ Wx_vec

        # Compute Wy and get compact vectorization
        Wy = Hy + Ht * self.DPhiY
        Wy_vec = Wy.view(np.float64).reshape(-1, 1)
        Wy_cvec = self.F2C_op @ Wy_vec

        # Solve for (X, Y) =  M \ (Wx, Wy)
        Wxy_cvec = np.vstack((Wx_cvec, Wy_cvec))
        out_XY = cho_solve(self.hess_fact, Wxy_cvec)
        out_XY = out_XY.reshape(2, -1)

        out_X = self.F2C_op.T @ out_XY[0]
        out_X = out_X.view(self.dtype).reshape((self.n, self.n))
        out[1][:] = (out_X + out_X.conj().T) * 0.5

        out_Y = self.F2C_op.T @ out_XY[1]
        out_Y = out_Y.view(self.dtype).reshape((self.n, self.n))
        out[2][:] = (out_Y + out_Y.conj().T) * 0.5

        # Solve for t = z^2 Ht + <DPhi(X, Y), (X, Y)>
        out_t = self.z2 * Ht
        out_t += inp(out_X, self.DPhiX)
        out_t += inp(out_Y, self.DPhiY)
        out[0][:] = out_t

        return out

    def invhess_congr(self, A):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.invhess_aux_updated:
            self.update_invhessprod_aux()
        if not self.congr_aux_updated:
            self.congr_aux(A)

        # The inverse Hessian product applied on (Ht, Hx, Hy) for the OPT
        # barrier is
        #     (X, Y) =  M \ (Wx, Wy)
        #         t  =  z^2 Ht + <DPhi(X, Y), (X, Y)>
        # where (Wx, Wy) = (Hx, Hy) + Ht DPhi(X, Y) and
        #     M = 1/z [ D2xxPhi D2xyPhi ] + [ X^1 ⊗ X^-1              ]
        #             [ D2yxPhi D2yyPhi ]   [              Y^1 ⊗ Y^-1 ]

        # Compute (Wx, Wy)
        np.outer(self.DPhi_cvec, self.At, out=self.work)
        self.work += self.Axy_cvec.T

        # Solve for (X, Y) =  M \ (Wx, Wy)
        out_xy = cho_solve(self.hess_fact, self.work)

        # Solve for t = z^2 Ht + <DPhi(X, Y), (X, Y)>
        out_t = self.z2 * self.At.reshape(-1, 1) + out_xy.T @ self.DPhi_cvec

        # Multiply A (H A')
        return x_dot_dense(self.Axy_cvec, out_xy) + np.outer(self.At, out_t)

    def third_dir_deriv_axpy(self, out, H, a=True):
        assert self.grad_updated
        if not self.hess_aux_updated:
            self.update_hessprod_aux()
        if not self.dder3_aux_updated:
            self.update_dder3_aux()

        (Ht, Hx, Hy) = H

        Dyxy, Dxyx, Uyxy, Uxyx = self.Dyxy, self.Dxyx, self.Uyxy, self.Uxyx
        D2yxy_h, D2yxy_xh = self.D2yxy_h, self.D2yxy_xh
        D2xyx_g, D2xyx_xg = self.D2xyx_g, self.D2xyx_xg
        UyxyYUyxy, UxyxXUxyx = self.UyxyYUyxy, self.UxyxXUxyx
        rt2Y_Uyxy, irt2Y_Uyxy = self.rt2Y_Uyxy, self.irt2Y_Uyxy
        rt2X_Uxyx, irt2X_Uxyx = self.rt2X_Uxyx, self.irt2X_Uxyx

        chi = Ht[0, 0] - inp(self.DPhiX, Hx) - inp(self.DPhiY, Hy)
        chi2 = chi * chi

        UyxyYHxYUyxy = irt2Y_Uyxy.conj().T @ Hx @ irt2Y_Uyxy
        UxyxXHyXUxyx = irt2X_Uxyx.conj().T @ Hy @ irt2X_Uxyx
        UxyxXHxXUxyx = irt2X_Uxyx.conj().T @ Hx @ irt2X_Uxyx
        UyxyYHyYUyxy = irt2Y_Uyxy.conj().T @ Hy @ irt2Y_Uyxy

        # Trace noncommutative perspective Hessians
        D2PhiXXH = scnd_frechet(D2yxy_h, UyxyYUyxy, UyxyYHxYUyxy, U=irt2Y_Uyxy)

        work = self.D1xyx_g * UxyxXHyXUxyx
        D2PhiXYH = irt2X_Uxyx @ work @ rt2X_Uxyx.conj().T
        D2PhiXYH += D2PhiXYH.conj().T
        D2PhiXYH -= scnd_frechet(D2xyx_xg, UxyxXUxyx, UxyxXHyXUxyx, U=irt2X_Uxyx)

        work = self.D1yxy_h * UyxyYHxYUyxy
        D2PhiYXH = irt2Y_Uyxy @ work @ rt2Y_Uyxy.conj().T
        D2PhiYXH += D2PhiYXH.conj().T
        D2PhiYXH -= scnd_frechet(D2yxy_xh, UyxyYUyxy, UyxyYHxYUyxy, U=irt2Y_Uyxy)

        D2PhiYYH = scnd_frechet(D2xyx_g, UxyxXUxyx, UxyxXHyXUxyx, U=irt2X_Uxyx)

        D2PhiXHH = inp(Hx, D2PhiXXH + D2PhiXYH)
        D2PhiYHH = inp(Hy, D2PhiYXH + D2PhiYYH)

        # Trace noncommutative perspective third order derivatives
        # Second derivatives of D_X trPg(X, Y)
        D3PhiXXX = thrd_frechet(Dyxy, D2yxy_h, self.d3h(Dyxy), irt2Y_Uyxy,
                                UyxyYUyxy, UyxyYHxYUyxy)  # fmt: skip

        work = rt2Y_Uyxy.conj().T @ Hy @ irt2Y_Uyxy
        work = work + work.conj().T
        D3PhiXXY = scnd_frechet(D2yxy_h, work, UyxyYHxYUyxy, U=irt2Y_Uyxy)
        D3PhiXXY -= thrd_frechet(Dyxy, D2yxy_xh, self.d3xh(Dyxy), irt2Y_Uyxy,
                                 UyxyYUyxy, UyxyYHxYUyxy, UyxyYHyYUyxy)  # fmt: skip
        D3PhiXYX = D3PhiXXY

        work = scnd_frechet(D2xyx_g, UxyxXHyXUxyx, UxyxXHyXUxyx, U=Uxyx)
        D3PhiXYY = self.irt2_X @ work @ self.rt2_X
        D3PhiXYY += D3PhiXYY.conj().T
        D3PhiXYY -= thrd_frechet(Dxyx, D2xyx_xg, self.d3xg(Dxyx), irt2X_Uxyx,
                                 UxyxXUxyx, UxyxXHyXUxyx)  # fmt: skip

        # Second derivatives of D_Y trPg(X, Y)
        D3PhiYYY = thrd_frechet(Dxyx, D2xyx_g, self.d3g(Dxyx), irt2X_Uxyx,
                                UxyxXUxyx, UxyxXHyXUxyx)  # fmt: skip

        work = rt2X_Uxyx.conj().T @ Hx @ irt2X_Uxyx
        work = work + work.conj().T
        D3PhiYYX = scnd_frechet(D2xyx_g, work, UxyxXHyXUxyx, U=irt2X_Uxyx)
        D3PhiYYX -= thrd_frechet(Dxyx, D2xyx_xg, self.d3xg(Dxyx), irt2X_Uxyx,
                                 UxyxXUxyx, UxyxXHyXUxyx, UxyxXHxXUxyx)  # fmt: skip
        D3PhiYXY = D3PhiYYX

        work = scnd_frechet(D2yxy_h, UyxyYHxYUyxy, UyxyYHxYUyxy, U=Uyxy)
        D3PhiYXX = self.irt2_Y @ work @ self.rt2_Y
        D3PhiYXX += D3PhiYXX.conj().T
        D3PhiYXX -= thrd_frechet(Dyxy, D2yxy_xh, self.d3xh(Dyxy), irt2Y_Uyxy,
                                 UyxyYUyxy, UyxyYHxYUyxy)  # fmt: skip

        # Third derivatives of barrier
        dder3_t = -2 * self.zi3 * chi2 - self.zi2 * (D2PhiXHH + D2PhiYHH)

        dder3_X = -dder3_t * self.DPhiX
        dder3_X -= 2 * self.zi2 * chi * (D2PhiXXH + D2PhiXYH)
        dder3_X += self.zi * (D3PhiXXX + D3PhiXXY + D3PhiXYX + D3PhiXYY)
        dder3_X -= 2 * self.inv_X @ Hx @ self.inv_X @ Hx @ self.inv_X
        dder3_X = (dder3_X + dder3_X.conj().T) * 0.5

        dder3_Y = -dder3_t * self.DPhiY
        dder3_Y -= 2 * self.zi2 * chi * (D2PhiYXH + D2PhiYYH)
        dder3_Y += self.zi * (D3PhiYYY + D3PhiYYX + D3PhiYXY + D3PhiYXX)
        dder3_Y -= 2 * self.inv_Y @ Hy @ self.inv_Y @ Hy @ self.inv_Y
        dder3_Y = (dder3_Y + dder3_Y.conj().T) * 0.5

        out[0][:] += dder3_t * a
        out[1][:] += dder3_X * a
        out[2][:] += dder3_Y * a

        return out

    # ========================================================================
    # Auxilliary functions
    # ========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        iscomplex = self.iscomplex

        # Get slices and views of A matrix to be used in congruence computations
        if sp.sparse.issparse(A):
            A = A.tocsr()
        self.Ax_vec = A[:, self.idx_X]
        self.Ay_vec = A[:, self.idx_Y]
        Ax_cvec = (self.F2C_op @ self.Ax_vec.T).T
        Ay_cvec = (self.F2C_op @ self.Ay_vec.T).T
        if sp.sparse.issparse(A):
            self.Axy_cvec = sp.sparse.hstack((Ax_cvec, Ay_cvec), format="coo")
        else:
            self.Axy_cvec = np.hstack((Ax_cvec, Ay_cvec))

        if sp.sparse.issparse(A):
            A = A.toarray()
        Ax_dense = np.ascontiguousarray(A[:, self.idx_X])
        Ay_dense = np.ascontiguousarray(A[:, self.idx_Y])
        self.At = A[:, 0]
        self.Ax = np.array([vec_to_mat(Ax_k, iscomplex) for Ax_k in Ax_dense])
        self.Ay = np.array([vec_to_mat(Ay_k, iscomplex) for Ay_k in Ay_dense])

        # Preallocate matrices we will need when performing these congruences
        self.work = np.empty_like(self.Axy_cvec.T)

        self.work0 = np.empty_like(self.Ax)
        self.work1 = np.empty_like(self.Ax)
        self.work2 = np.empty_like(self.Ax)
        self.work3 = np.empty_like(self.Ax)
        self.work4 = np.empty_like(self.Ax)
        self.work5 = np.empty_like(self.Ax)

        self.work6 = np.empty((self.Ax.shape[::-1]), dtype=self.dtype)

        self.congr_aux_updated = True

    def update_hessprod_aux(self):
        assert not self.hess_aux_updated
        assert self.grad_updated

        Dyxy, Dxyx = self.Dyxy, self.Dxyx

        self.D1yxy_xh = D1_f(Dyxy, self.xh(Dyxy), self.dxh(Dyxy))
        self.D1xyx_xg = D1_f(Dxyx, self.xg(Dxyx), self.dxg(Dxyx))

        self.D2yxy_h = D2_f(Dyxy, self.D1yxy_h, self.d2h(Dyxy))
        self.D2xyx_g = D2_f(Dxyx, self.D1xyx_g, self.d2g(Dxyx))
        self.D2yxy_xh = D2_f(Dyxy, self.D1yxy_xh, self.d2xh(Dyxy))
        self.D2xyx_xg = D2_f(Dxyx, self.D1xyx_xg, self.d2xg(Dxyx))

        # Preparing other required variables
        self.zi2 = self.zi * self.zi

        self.hess_aux_updated = True

    def update_invhessprod_aux(self):
        assert not self.invhess_aux_updated
        assert self.grad_updated
        assert self.hess_aux_updated
        if not self.invhess_aux_aux_updated:
            self.update_invhessprod_aux_aux()

        # Precompute and factorize the matrix
        #     M = 1/z [ D2xxPhi D2xyPhi ] + [ X^1 ⊗ X^-1               ]
        #             [ D2yxPhi D2yyPhi ]   [               Y^1 ⊗ Y^-1 ]

        self.z2 = self.z * self.z

        D2xyx_g, D2xyx_xg = self.D2xyx_g, self.D2xyx_xg
        D2yxy_h, irt2Y_Uyxy = self.D2yxy_h, self.irt2Y_Uyxy
        UyxyYUyxy, UxyxXUxyx = self.UyxyYUyxy, self.UxyxXUxyx
        rt2X_Uxyx, irt2X_Uxyx = self.rt2X_Uxyx, self.irt2X_Uxyx

        work10, work11 = self.work10, self.work11
        work12, work13, work14 = self.work12, self.work13, self.work14

        # ======================================================================
        # Construct XX block of Hessian, i.e., (D2xxPhi + X^1 ⊗ X^-1)
        # ======================================================================
        # D2_XX trPg(X, Y)[Hx] = Y^-½ D2h(Y^-½ X Y^½)[Y, Y^-½ Hx Y^-½] Y^-½
        congr_multi(work14, irt2Y_Uyxy.conj().T, self.E, work=work13)
        scnd_frechet_multi(work11, D2yxy_h, work14, UyxyYUyxy, U=irt2Y_Uyxy,
                           work1=work12, work2=work13, work3=work10)  # fmt: skip
        work11 *= self.zi
        # X^1 Eij X^-1
        congr_multi(work14, self.inv_X, self.E, work=work13)
        work14 += work11
        # Vectorize matrices as compact vectors to get square matrix
        work = work14.view(np.float64).reshape((self.vn, -1))
        Hxx = x_dot_dense(self.F2C_op, work.T)

        # ======================================================================
        # Construct YY block of Hessian, i.e., (D2yyPhi + Y^1 ⊗ Y^-1)
        # ======================================================================
        # D2_YY trPg(X, Y)[Hy] = X^-½ D2g(X^-½ Y X^½)[X, X^-½ Hy X^-½] X^-½
        congr_multi(work14, irt2X_Uxyx.conj().T, self.E, work=work13)
        scnd_frechet_multi(work11, D2xyx_g, work14, UxyxXUxyx, U=irt2X_Uxyx,
                           work1=work12, work2=work13, work3=work10)  # fmt: skip
        work11 *= self.zi
        # Y^1 Eij Y^-1
        congr_multi(work12, self.inv_Y, self.E, work=work13)
        work12 += work11
        # Vectorize matrices as compact vectors to get square matrix
        work = work12.view(np.float64).reshape((self.vn, -1))
        Hyy = x_dot_dense(self.F2C_op, work.T)

        # ======================================================================
        # Construct XY block of Hessian, i.e., D2yxPhi
        # ======================================================================
        # D2_XY trPg(X, Y)[Hy]
        #     = -X^-½ D2xg(X^-½ Y X^-½)[X, X^-½ Hy X^-½] X^-½
        #       + X^½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^-½
        #       + X^-½ Dg(X^-½ Y X-^½)[X^-½ Hy X^-½] X^½
        # First term, i.e., -X^-½ D2xg(X^-½ Y X^-½)[X, X^-½ Hy X^-½] X^-½
        scnd_frechet_multi(work11, D2xyx_xg, work14, UxyxXUxyx, U=irt2X_Uxyx,
                           work1=work12, work2=work13, work3=work10)  # fmt: skip
        # Second and third terms, i.e., X^½ [ ... ] X^-½ + X^-½ [ ... ] X^½
        work14 *= self.D1xyx_g
        congr_multi(work12, irt2X_Uxyx, work14, work=work13, B=rt2X_Uxyx)
        np.add(work12, work12.conj().transpose(0, 2, 1), out=work13)
        work13 -= work11
        work13 *= self.zi
        # Vectorize matrices as compact vectors to get square matrix
        work = work13.view(np.float64).reshape((self.vn, -1))
        Hxy = x_dot_dense(self.F2C_op, work.T)

        # Construct Hessian and factorize
        Hxx = (Hxx + Hxx.conj().T) * 0.5
        Hyy = (Hyy + Hyy.conj().T) * 0.5

        self.hess[: self.vn, : self.vn] = Hxx
        self.hess[self.vn :, self.vn :] = Hyy
        self.hess[self.vn :, : self.vn] = Hxy.T
        self.hess[: self.vn, self.vn :] = Hxy

        self.hess_fact = cho_fact(self.hess)
        self.invhess_aux_updated = True

        return

    def update_invhessprod_aux_aux(self):
        assert not self.invhess_aux_aux_updated

        self.precompute_computational_basis()

        self.hess = np.empty((2 * self.vn, 2 * self.vn))

        self.work10 = np.empty((self.n, self.n, self.vn), dtype=self.dtype)
        self.work11 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)
        self.work12 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)
        self.work13 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)
        self.work14 = np.empty((self.vn, self.n, self.n), dtype=self.dtype)

        self.invhess_aux_aux_updated = True

    def update_dder3_aux(self):
        assert not self.dder3_aux_updated
        assert self.hess_aux_updated

        self.zi3 = self.zi2 * self.zi

        self.dder3_aux_updated = True

    def get_central_ray(self):
        # Solve a 3-dimensional nonlinear system of equations to get the central
        # point of the barrier function
        n = self.n
        (t, x, y) = (1.0 + n * self.g(1.0), 1.0, 1.0)

        for _ in range(10):
            # Precompute some useful things
            z = t - n * x * self.g(y / x)
            zi = 1 / z
            zi2 = zi * zi

            dx = self.dh(x / y)
            dy = self.dg(y / x)

            d2dx2 = self.d2h(x / y) / y
            d2dy2 = self.d2g(y / x) / x
            d2dxdy = -d2dy2 * y / x

            # Get gradient
            g = np.array([t - zi, 
                          n * x + n * dx * zi - n / x, 
                          n * y + n * dy * zi - n / y])  # fmt: skip

            # Get Hessian
            (Htt, Htx, Hty) = (zi2, -n * zi2 * dx, -n * zi2 * dy)
            Hxx = n * n * zi2 * dx * dx + n * zi * d2dx2 + n / x / x
            Hyy = n * n * zi2 * dy * dy + n * zi * d2dy2 + n / y / y
            Hxy = n * n * zi2 * dx * dy + n * zi * d2dxdy

            H = np.array([[Htt + 1, Htx, Hty],
                          [Htx, Hxx + n, Hxy],
                          [Hty, Hxy, Hyy + n]])  # fmt: skip

            # Perform Newton step
            delta = -np.linalg.solve(H, g)
            decrement = -np.dot(delta, g)

            # Check feasible
            (t1, x1, y1) = (t + delta[0], x + delta[1], y + delta[2])
            if x1 < 0 or y1 < 0 or t1 < n * x1 * self.g(y1 / x1):
                # Exit if not feasible and return last feasible point
                break

            (t, x, y) = (t1, x1, y1)

            # Exit if decrement is small, i.e., near optimality
            if decrement / 2.0 <= 1e-12:
                break

        return (t, x, y)



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import itertools

import numba as nb
import numpy as np
import scipy as sp

from qics._utils.linalg import congr_multi, scale_axis, x_dot_dense
from qics.cones.base import SymCone
from qics.vectorize import vec_to_mat


class PosSemidefinite(SymCone):
    r"""A class representing a positive semidefinite cone defined either
    on real symmetric matrices

    .. math::

        \mathbb{S}^n_+ = \{ X \in \mathbb{S}^n : X \succeq 0 \},

    or complex Hermitian matrices

    .. math::

        \mathbb{H}^n_+ = \{ X \in \mathbb{H}^n : X \succeq 0 \}.

    Parameters
    ----------
    n : :obj:`int`
        Dimension of the matrix :math:`X`.
    iscomplex : :obj:`bool`
        Whether the matrix :math:`X` is defined over :math:`\mathbb{H}^n`
        (``True``), or restricted to :math:`\mathbb{S}^n` (``False``). The
        default is ``False``.
    """

    def __init__(self, n, iscomplex=False):
        self.n = n
        self.iscomplex = iscomplex

        self.nu = n  # Barrier parameter

        if iscomplex:
            self.dim = [2 * n * n]
            self.type = ["h"]
            self.dtype = np.complex128
        else:
            self.dim = [n * n]
            self.type = ["s"]
            self.dtype = np.float64

        # Get LAPACK operators
        X = np.eye(self.n, dtype=self.dtype)
        self.cho_fact = sp.linalg.lapack.get_lapack_funcs("potrf", (X,))
        self.cho_inv = sp.linalg.lapack.get_lapack_funcs("trtri", (X,))
        self.svd = sp.linalg.lapack.get_lapack_funcs("gesdd", (X,))
        self.svd_lwork = sp.linalg.lapack.get_lapack_funcs("gesdd_lwork", (X,))(n, n)
        if iscomplex:
            self.eigvalsh = sp.linalg.lapack.get_lapack_funcs("heevr", (X,))
        else:
            self.eigvalsh = sp.linalg.lapack.get_lapack_funcs("syevr", (X,))

        # Update flags
        self.feas_updated = False
        self.grad_updated = False
        self.nt_aux_updated = False
        self.congr_aux_updated = False

        return

    def get_iscomplex(self):
        return self.iscomplex

    def get_init_point(self, out):
        point = [np.eye(self.n, dtype=self.dtype)]

        self.set_point(point, point)

        out[0][:] = self.X
        return out

    def set_point(self, primal, dual=None, a=True):
        self.X = primal[0] * a
        self.Z = dual[0] * a if (dual is not None) else None

        self.feas_updated = False
        self.grad_updated = False
        self.nt_aux_updated = False

    def set_dual(self, dual, a=True):
        self.Z = dual[0] * a

    def get_feas(self):
        if self.feas_updated:
            return self.feas

        self.feas_updated = True

        # Check that X is PSD by trying to Cholesky factor it
        self.X_chol, info = self.cho_fact(self.X, lower=True)
        if info != 0:
            self.feas = False
            return self.feas

        # Check that Z is PSD by trying to Cholesky factor it
        if self.Z is not None:
            self.Z_chol, info = self.cho_fact(self.Z, lower=True)
            if info != 0:
                self.feas = False
                return self.feas

        self.feas = True
        return self.feas

    def get_dual_feas(self):
        self.Z_chol, info = self.cho_fact(self.Z, lower=True)
        return info == 0

    def get_val(self):
        (sign, logabsdet) = np.linalg.slogdet(self.X)
        return -sign * logabsdet

    def update_grad(self):
        assert not self.grad_updated

        self.X_chol_inv, _ = self.cho_inv(self.X_chol, lower=True)
        self.X_inv = self.X_chol_inv.conj().T @ self.X_chol_inv
        self.grad = [-self.X_inv]

        self.grad_updated = True

    def hess_prod_ip(self, out, H):
        if not self.grad_updated:
            self.update_grad()
        XHX = self.X_inv @ H[0] @ self.X_inv
        np.add(XHX, XHX.conj().T, out=out[0])
        out[0] *= 0.5
        return out

    def hess_congr(self, A):
        if not self.grad_updated:
            self.update_grad()
        return self.base_congr(A, self.X_inv, self.X_chol_inv.conj().T)

    def invhess_prod_ip(self, out, H):
        XHX = self.X @ H[0] @ self.X
        np.add(XHX, XHX.conj().T, out=out[0])
        out[0] *= 0.5
        return out

    def invhess_congr(self, A):
        return self.base_congr(A, self.X, self.X_chol)

    def base_congr(self, A, X, X_rt2):
        # Generalized function to compute the matrix [<Ai, X Aj X>]_ij for a
        # a given matrix X
        if not self.congr_aux_updated:
            self.congr_aux(A)

        (n, p, p_ds) = (self.n, A.shape[0], len(self.A_ds_idxs))
        out = np.zeros((p, p))

        if len(self.A_sp_idxs) > 0:
            # Compute sparse-sparse component using Numba compiled functions
            if self.iscomplex:
                _sparse_congr_complex(out, self.A_sp_rows, self.A_sp_cols, 
                                      self.A_sp_data, self.A_sp_nnzs, X, self.A_sp_idxs)  # fmt: skip
            else:
                _sparse_congr(out, self.A_sp_rows, self.A_sp_cols, 
                              self.A_sp_data, self.A_sp_nnzs, X, self.A_sp_idxs)  # fmt: skip

            # Compute sparse-dense and dense-dense components
            if p_ds > 0:
                work = self.work

                # Compute X Aj X for all dense Aj
                lhs = np.zeros((p_ds, self.dim[0]))
                lhs_view = lhs.reshape((p_ds, n, -1)).view(self.dtype)
                congr_multi(lhs_view, X, self.Ai_ds, work=work)

                # Compute inner products <Ai, X Aj X> for all dense Aj
                temp = x_dot_dense(self.A_triu, lhs[:, self.triu_idxs].T)
                out[:, self.A_ds_idxs] = temp
                out[self.A_ds_idxs, :] = temp.T
        else:
            # If there are no (nonzero) sparse Aj, then compute congruence using
            # the symmetric multiplication [A (L' kr L)] [A (L' kr L)]'
            work = self.work

            # Compute L' Aj L
            lhs = np.zeros((p_ds, self.dim[0]))
            lhs_view = lhs.reshape((p_ds, n, -1)).view(self.dtype)
            congr_multi(lhs_view, X_rt2.conj().T, self.Ai_ds, work=work)

            # Compute inner products <L' Ai L, L' Aj L>
            out[self.A_ds_ds_idxs] = lhs @ lhs.T

        return out

    def third_dir_deriv_axpy(self, out, H, a=True):
        if not self.grad_updated:
            self.update_grad()

        XHX_2 = self.X_inv @ H[0] @ self.X_chol_inv.conj().T
        out[0] -= 2 * a * XHX_2 @ XHX_2.conj().T
        return out

    def prox(self):
        assert self.feas_updated
        XZX_I = self.X_chol.conj().T @ self.Z @ self.X_chol
        XZX_I.flat[:: self.n + 1] -= 1
        return np.linalg.norm(XZX_I) ** 2

    # ==========================================================================
    # Functions specific to symmetric cones for NT scaling
    # ==========================================================================
    # Computes the NT scaling point W and scaled variable Lambda such that
    #     H(W)[S] = Z  <==> Lambda := P^-T(S) = P(Z)
    # where H(W) = V^T V. To obtain for for the PSD cone, first let compute the
    # SVD
    #     U D V^T = Z_chol^T S_chol
    # Then compute
    #     R    := S_chol V D^-1/2     = Z_chol^-T U D^1/2
    #     R^-1 := D^1/2 V^T S_chol^-1 = D^-1/2 U^T Z_chol^T
    # Then we can find the scaling point as
    #     W    := R R^T
    #           = S^1/2 (S^1/2 Z S^1/2)^-1/2 S^1/2
    #           = Z^-1/2 (Z^1/2 S Z^1/2)^1/2 Z^1/2 (i.e., geo. mean of Z and S)
    #     W^-1 := R^-T R^-1
    # and the scaled point as
    #     Lambda := D
    # Also, we have the linear transformations given by
    #     H(W)[S] = W^-1 S W^-1
    #     P^-T(S) = R^-1 S R^-T
    #     P(Z)    = R^T Z R
    # See: [Section 4.3]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

    def nt_aux(self):
        assert not self.nt_aux_updated
        # Take the SVD of Z_chol^T S_chol to get scaled point Lambda := D
        RL = self.Z_chol.conj().T @ self.X_chol
        _, self.Lambda, Vt, _ = self.svd(RL, lwork=self.svd_lwork)
        D_rt2 = np.sqrt(self.Lambda)

        # Compute the scaling point as
        #    R := S_chol V D^-1/2, and
        #    W := R R^T
        self.R = self.X_chol @ (Vt.conj().T / D_rt2)
        self.W = self.R @ self.R.conj().T

        # Compute the inverse scaling point as
        #    R^-1 := D^1/2 V^T S_chol^-1, and
        #    W^-1 := R^-T R^-1
        self.X_chol_inv, _ = self.cho_inv(self.X_chol, lower=True)
        self.R_inv = (self.X_chol_inv.conj().T @ (Vt.conj().T * D_rt2)).conj().T
        self.W_inv = self.R_inv.conj().T @ self.R_inv

        self.nt_aux_updated = True

    def nt_prod_ip(self, out, H):
        if not self.nt_aux_updated:
            self.nt_aux()
        WHW = self.W_inv @ H[0] @ self.W_inv
        np.add(WHW, WHW.conj().T, out=out[0])
        out[0] *= 0.5

    def nt_congr(self, A):
        if not self.nt_aux_updated:
            self.nt_aux()
        return self.base_congr(A, self.W_inv, self.R_inv.conj().T)

    def invnt_prod_ip(self, out, H):
        if not self.nt_aux_updated:
            self.nt_aux()
        WHW = self.W @ H[0] @ self.W
        np.add(WHW, WHW.conj().T, out=out[0])
        out[0] *= 0.5

    def invnt_congr(self, A):
        if not self.nt_aux_updated:
            self.nt_aux()
        return self.base_congr(A, self.W, self.R)

    def comb_dir(self, out, dS, dZ, sigma_mu):
        # Compute the residual for rs where rs is given as the lhs of
        #     Lambda o (W dz + W^-T ds) = -Lambda o Lambda - (W^-T ds_a) o (W dz_a)
        #                                 + sigma * mu * I
        # which is rearranged into the form H ds + dz = rs, i.e.,
        #     rs := W^-1 [ Lambda \ (-Lambda o Lambda - (W^-T ds_a) o (W dz_a) + sigma*mu I) ]
        # See: [Section 5.4]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

        if not self.nt_aux_updated:
            self.nt_aux()

        # Compute (W^-T ds_a) o (W dz_a) = (R^-1 dS R^-T) o (R^T dZ R)
        #               = 0.5 * ([R^-1 dS dZ R] + [R^T dZ dS R^-T])
        temp1 = self.R_inv @ dS[0]
        temp2 = dZ[0] @ self.R
        temp3 = temp1 @ temp2
        np.add(temp3, temp3.conj().T, out=temp1)
        temp1 *= -0.5

        # Compute -Lambda o Lambda - [ ... ] + sigma*mu I
        # Note that Lambda is a diagonal matrix
        temp1.flat[:: self.n + 1] -= np.square(self.Lambda)
        temp1.flat[:: self.n + 1] += sigma_mu

        # Compute Lambda \ [ ... ]
        # Since Lambda is diagonal, the solution to the Sylvester equation
        #     find  X  s.t.  0.5 * (Lambda X + X Lambda) = B
        # is given by
        #     X = B .* (2 / [Lambda_ii + Lambda_jj]_ij)
        Gamma = np.add.outer(self.Lambda, self.Lambda)
        temp1 /= Gamma

        # Compute W^-1 [ ... ] = R^-T [... ] R^-1
        temp = self.R_inv.conj().T @ temp1 @ self.R_inv
        np.add(temp, temp.conj().T, out=out[0])

    def step_to_boundary(self, dS, dZ):
        # Compute the maximum step alpha in [0, 1] we can take such that
        #     S + alpha dS >= 0
        #     Z + alpha dZ >= 0
        # See: [Section 8.3]https://www.seas.ucla.edu/~vandenbe/publications/coneprog.pdf

        if not self.nt_aux_updated:
            self.nt_aux()
        Lambda_irt2 = np.reciprocal(np.sqrt(self.Lambda))

        # Compute rho := H(lambda)^1/2 W^-T dS
        #              = Lambda^-1/2 R^-1 dS R^-T Lambda^-1/2
        rho = self.R_inv @ dS @ self.R_inv.conj().T
        rho *= Lambda_irt2.reshape((-1, 1))
        rho *= Lambda_irt2.reshape((1, -1))

        # Compute sig := H(lambda)^1/2 W dS
        #              = Lambda^-1/2 R^T dS R Lambda^-1/2
        sig = self.R.conj().T @ dZ @ self.R
        sig *= Lambda_irt2.reshape((-1, 1))
        sig *= Lambda_irt2.reshape((1, -1))

        # Compute minimum eigenvalues of rho and sig
        min_eig_rho = self.eigvalsh(rho, compute_v=False, range="I", iu=1)[0][0]
        min_eig_sig = self.eigvalsh(sig, compute_v=False, range="I", iu=1)[0][0]

        # Maximum step is given by
        #     alpha := 1 / max(0, -min(eig(rho)), -min(eig(sig)))
        # Clamp this step between 0 and 1
        if min_eig_rho >= 0 and min_eig_sig >= 0:
            return 1.0
        else:
            return 1.0 / max(-min_eig_rho, -min_eig_sig)

    # ==========================================================================
    # Auxilliary functions
    # ==========================================================================
    def congr_aux(self, A):
        assert not self.congr_aux_updated

        n = self.n

        if sp.sparse.issparse(A):
            A = A.tocsr()

            # Split A into sparse and dense groups
            A_nnz = A.getnnz(1)
            self.A_sp_idxs = np.where((A_nnz > 0) & (A_nnz < n))[0]
            self.A_ds_idxs = np.where((A_nnz >= n))[0]

            self.A_sp_idxs = self.A_sp_idxs[np.argsort(A_nnz[self.A_sp_idxs])]
            self.A_ds_ds_idxs = np.ix_(self.A_ds_idxs, self.A_ds_idxs)

            # Prepare things we need for Strategy 1, i.e., sparse-sparse
            # congruence using Numba functions
            if len(self.A_sp_idxs) > 0:
                A_sp = A[self.A_sp_idxs]

                triu_idxs = _get_triu_idxs(n)

                if self.iscomplex:
                    A_sp_real = A_sp[:, ::2][:, triu_idxs]
                    A_sp_imag = A_sp[:, 1::2][:, triu_idxs]
                    A_sp_lil = (A_sp_real + A_sp_imag * 1j).tolil()
                else:
                    A_sp_lil = A_sp[:, triu_idxs].tolil()

                # Get number of nonzeros for each sparse Ai (to account for
                # ragged arrays)
                self.A_sp_nnzs = A_sp_lil.getnnz(1)

                # Get rows and columns of nonzeros of sparse Ai
                rowcols = _lil_to_array(A_sp_lil.rows)
                self.A_sp_rows, self.A_sp_cols = _triu_idx_to_ij(rowcols)

                # Get values of nonzeros of sparse Ai, and scale off-diagonal
                # elements to account for only using upper triangular nonzeros
                self.A_sp_data = _lil_to_array(A_sp_lil.data)
                self.A_sp_data[self.A_sp_cols != self.A_sp_rows] *= 2

            # Prepare things we need for Strategy 2, i.e., sparse-dense and
            # dense-dense congruence by computing X Aj X for all dense Aj, then
            # <Ai, X Aj X> for all Ai and dense Aj
            if len(self.A_ds_idxs) > 0:
                A_ds = A[self.A_ds_idxs, :]

                # Turn rows of A into matrices Ai
                if self.iscomplex:
                    A_ds = A_ds[:, ::2] + A_ds[:, 1::2] * 1j
                A_ds = A_ds.toarray()

                self.Ai_ds = np.array([Ai.reshape((n, n)) for Ai in A_ds])
                self.work = np.zeros_like(self.Ai_ds)

                # Get upper triangular slices of A so we can more efficiently
                # do the inner product <Ai, X Aj X> as a matrix multiplication
                if len(self.A_sp_idxs) > 0:
                    # Scale upper triangular elements by 2 to account for only
                    # using upper triangular elements
                    scale = 2 * np.ones(self.dim[0])
                    if self.iscomplex:
                        scale[:: 2 * n + 2] = 1
                    else:
                        scale[:: n + 1] = 1
                    self.triu_idxs = _get_triu_idxs(n, self.iscomplex)
                    self.A_triu = scale_axis(A, scale_cols=scale).tocsr()
                    self.A_triu = self.A_triu[:, self.triu_idxs].tocoo()

        else:
            # A and all Ai are dense matrices
            # Just need to convert the rows of A into dense matrices
            A = np.ascontiguousarray(A)

            self.A_sp_idxs = np.array([])
            self.A_ds_idxs = np.arange(A.shape[0])
            self.A_ds_ds_idxs = np.ix_(self.A_ds_idxs, self.A_ds_idxs)

            self.Ai_ds = np.array([vec_to_mat(Ak, self.iscomplex) for Ak in A])
            self.work = np.zeros_like(self.Ai_ds)

        self.congr_aux_updated = True


def _lil_to_array(ragged):
    # Converts a list of lists (with possibly different lengths) into a numpy
    # array padded with zeros
    padded_list = list(itertools.zip_longest(*ragged, fillvalue=0))
    return np.array(padded_list).T


def _triu_idx_to_ij(idx):
    # Converts upper triangular indices to (i,j) coordinates
    #     [ 0  1  3       ]         [ (0,0)  (0,1)  (0,2)       ]
    #     [    2  4  ...  ]   -->   [        (1,1)  (1,2)  ...  ]
    #     [       5       ]         [               (2,2)       ]
    # See: https://stackoverflow.com/questions/40950460/
    j = np.ceil(np.sqrt(2 * (idx + 1) + 0.25) - 0.5) - 1
    i = idx - (j + 1) * j / 2
    return i.astype("int32"), j.astype("int32")


def _get_triu_idxs(n, iscomplex=False):
    # Gets indices of a vectorized matrix corresponding to the upper triangular
    # elements of the matrix
    if iscomplex:
        diag = [2 * (i + i * n) for i in range(n)]
        triu_real = [2 * (j + i * n) for j in range(n) for i in range(j)]
        triu_imag = [2 * (j + i * n) + 1 for j in range(n) for i in range(j)]
        return np.array(diag + triu_real + triu_imag)
    else:
        return np.array([j + i * n for j in range(n) for i in range(j + 1)])


# ============================================================================
# Numba functions for computing Schur complement matrix when A is very sparse
# ============================================================================


@nb.njit(cache=True, parallel=True, fastmath=True)
def _sparse_congr(out, A_rows, A_cols, A_vals, A_nnz, X, indices):
    # Computes the congruence transform A (X kron X) A' when A is very sparse
    # See https://link.springer.com/article/10.1007/BF02614319

    # We can cut the amount of operations in half by exploiting symetry of A and
    # X as follows
    # (AHA)_ij = Σ_a,b (Ai)_ab (Σ_c,d (Aj)_cd X_ac X_db)
    #          = Σ_a,b (Ai)_ab (  [Σ_c=d (Aj)_cd X_ac X_db]
    #                             + [Σ_c<d (Aj)_cd X_ac X_db]
    #                             + [Σ_c>d (Aj)_cd X_ac X_db]  )
    #          = Σ_a,b (Ai)_ab (  [Σ_c=d (Aj)_cd X_ac X_db]
    #                             + [Σ_c<d (Aj)_cd (X_ac X_db + X_ad X_cb)]  )
    #          = [Σ_a=b (Ai)_ab ( ... )] + [Σ_a<b (Ai)_ab ( ... )] + [Σ_a>b (Ai)_ab ( ... )]
    #          = [Σ_a=b (Ai)_ab ( ... )] + 2 [Σ_a<b (Ai)_ab ( ... )]
    # Note that we assume off-diagonal entries of Ai have been scaled by 2.
    # Also note that we assume only upper triangular elements are given to us so
    # c < d.

    n = A_rows.shape[0]

    # Loop through upper triangular entries of the Schur complement matrix (AHA)_ij
    for j in nb.prange(n):
        for i in nb.prange(j + 1):
            i_AHA = indices[i]
            j_AHA = indices[j]

            tmp1 = 0.0

            # Loop over nonzero entries of Ai
            for alpha in range(A_nnz[i]):
                a = A_rows[i, alpha]
                b = A_cols[i, alpha]

                tmp2 = 0.0
                tmp3 = 0.0

                # Loop over nonzero entries of Aj
                for beta in range(A_nnz[j]):
                    c = A_rows[j, beta]
                    d = A_cols[j, beta]

                    if c < d:
                        tmp2 += A_vals[j, beta] * (
                            X[a, c] * X[d, b] + X[a, d] * X[c, b]
                        )
                    else:
                        tmp3 += A_vals[j, beta] * X[a, c] * X[d, b]

                tmp1 += A_vals[i, alpha] * (0.5 * tmp2 + tmp3)

            if i_AHA <= j_AHA:
                out[i_AHA, j_AHA] = tmp1
            else:
                out[j_AHA, i_AHA] = tmp1


@nb.njit(cache=True, parallel=True, fastmath=True)
def _sparse_congr_complex(out, A_rows, A_cols, A_vals, A_nnz, X, indices):
    # Computes the congruence transform A (X kron X) A' when A is very sparse
    # See https://link.springer.com/article/10.1007/BF02614319

    # We can cut the amount of operations in half by exploiting symetry of A and
    # X as follows
    # (AHA)_ij = Σ_a,b (Ai)_ab* (Σ_c,d (Aj)_cd X_ac X_db)
    #          = Σ_a,b (Ai)_ab* (  [Σ_c=d (Aj)_cd X_ac X_db]
    #                              + [Σ_c>d (Aj)_cd X_ac X_db]
    #                              + [Σ_c<d (Aj)_cd X_ac X_db]  )
    #          = Σ_a,b (Ai)_ab* (  [Σ_c=d (Aj)_cd X_ac X_db]
    #                              + [Σ_c<d (Aj)_cd X_ac X_db + (Aj)_cd* X_ad X_cb]  )
    #          = [Σ_a=b (Ai)_ab ( ... )] + [Σ_a<b (Ai)_ab ( ... )] + [Σ_a>b (Ai)_ab ( ... )]
    #          = [Σ_a=b (Ai)_ab ( ... )] + [Σ_a>b (Ai)_ab ( ... ) + (Ai)_ab* ( ... )]
    # Also note that off-diagonal entries of Ai have been scaled by 2

    n = A_rows.shape[0]

    # Loop through each entry of the Schur complement matrix (AHA)_ij
    for j in nb.prange(n):
        for i in nb.prange(j + 1):
            i_AHA = indices[i]
            j_AHA = indices[j]

            tmp1 = 0.0
            for alpha in range(A_nnz[i]):
                a = A_rows[i, alpha]
                b = A_cols[i, alpha]

                tmp2 = 0.0
                tmp3 = 0.0
                for beta in range(A_nnz[j]):
                    c = A_rows[j, beta]
                    d = A_cols[j, beta]

                    if c < d:
                        tmp2 += A_vals[j, beta] * X[a, c] * X[d, b]
                        tmp2 += np.conj(A_vals[j, beta]) * X[a, d] * X[c, b]
                    else:
                        tmp3 += A_vals[j, beta].real * X[a, c] * X[d, b]

                # Do addition slightly differently to guarantee a real number
                # i.e., just take the inner product between Ai and X Aj X by
                #     Σ_ab Re[(X Aj X)_ab] * Re[(Ai)_ab] + 2 Im[(X Aj X)_ab] * Im[(Ai)_ab]
                tmp3 += 0.5 * tmp2
                tmp1 += A_vals[i, alpha].real * tmp3.real
                tmp1 += A_vals[i, alpha].imag * tmp3.imag

            if i_AHA <= j_AHA:
                out[i_AHA, j_AHA] = tmp1
            else:
                out[j_AHA, i_AHA] = tmp1



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import numpy as np
import scipy as sp

from qics._utils.linalg import cho_fact, cho_solve
from qics.point import Point, PointXYZ, VecProduct


class KKTSolver:
    """A class which is used to solve the KKT system

        [ rx ]    [      A'  G'  c ] [ dx ]   [    ]
        [ ry ] := [ -A           b ] [ dy ] - [    ]
        [ rz ]    [ -G           h ] [ dz ]   [ ds ]
        [rtau]    [ -c' -b' -h'    ] [dtau]   [dkap]

    and

        rs   :=  mu H(s) ds + dz
        rkap :=  (mu / tau^2) dtau + dkap

    or, if NT scaling is used,

        rs   :=  H(w) ds + dz
        rkap :=  (kap / tau) dtau + dkap

    for (dx, dy, dz, dtau, dkap) given right-hand residuals (rx, ry, rz, rtau, rkap) by
    using block elimination and Cholesky factorization of the Schur complement matrix.
    """

    def __init__(self, model, ir=True, use_invhess=True):
        self.model = model

        # Iterative refinement settings
        self.ir = ir  # Use iterative refinement or not
        self.ir_settings = {
            "maxiter": 1,  # Maximum IR iterations
            "tol": 1e-8,  # Tolerance for when IR is used
            "improv_ratio": 5.0,  # Expected reduction in tolerance, else slow progress
        }

        self.use_invhess = use_invhess

        # Preallocate vectors do to computations with
        self.cbh = PointXYZ(model)
        self.cbh.x[:] = model.c
        self.cbh.y[:] = model.b
        self.cbh.z.vec[:] = model.h

        self.ir_pnt = Point(model)
        self.res_pnt = Point(model)

        self.ir_xyz = PointXYZ(model)
        self.res_xyz = PointXYZ(model)
        self.c_xyz = PointXYZ(model)
        self.v_xyz = PointXYZ(model)

        self.work1 = VecProduct(model.cones)
        self.work2 = VecProduct(model.cones)

        self.GHG_fact = None

        return

    def update_lhs(self, model, point, mu):
        self.mu = mu
        self.pnt = point

        # Precompute and factor Schur complement matrix
        if model.use_G:
            if self.GHG_fact is None or self.use_invhess or model.issymmetric:
                GHG = blk_hess_congruence(model.G_T_views, model)
                self.GHG_fact = cho_fact(GHG, increment_diag=(not model.use_A))

            if model.use_A:
                self.GHG_issingular = self.GHG_fact is None
                if self.GHG_issingular:
                    # GHG is singular, Cholesky factor GHG + AA instead
                    GHG += model.A.T @ model.A
                    self.GHG_fact = cho_fact(GHG)

                GHGA = cho_solve(self.GHG_fact, model.A_T_dense)
                AGHGA = model.A @ GHGA
                self.AGHGA_fact = cho_fact(AGHGA)

        elif model.use_A:
            AHA = blk_invhess_congruence(model.A_invG_views, model)
            self.AHA_fact = cho_fact(AHA)

        # Solve constant 3x3 subsystem
        self.solve_sys_3(self.c_xyz, self.cbh)
        if self.ir and self.use_invhess:
            self.solve_sys_3_ir(self.c_xyz, self.cbh)

        return

    def solve_sys(self, dir, rhs, ir=True):
        # Solve system
        self.solve_sys_6(dir, rhs)
        if self.ir and ir:
            res_norm = self.solve_sys_6_ir(dir, rhs)
        else:
            res_norm = 0.0

        return res_norm

    def solve_sys_6(self, d, r):
        # Compute (dx, dy, dz, ds, dtau, dkap) by solving the
        # 6x6 block system
        #     [ rx ]    [      A'  G'  c ]  [ dx ]   [    ]
        #     [ ry ] := [ -A           b ]  [ dy ] - [    ]
        #     [ rz ]    [ -G           h ]  [ dz ]   [ ds ]
        #     [rtau]    [ -c' -b' -h'    ]  [dtau]   [dkap]
        # and
        #       rs   :=  mu H(s) ds + dz
        #      rkap  :=  (mu / tau^2) dtau + dkap
        # or, if NT scaling is used,
        #       rs   :=  H(w) ds + dz
        #      rkap  :=  (kap / tau) dtau + dkap
        # for a given (rx, ry, rz, rs, rtau, rkap).

        model = self.model
        pnt = self.pnt

        # First, solve the two reduced 3x3 subsystems
        #     [ rx ]    [      A'  G' ]  [ dx ]
        #     [ ry ] := [ -A          ]  [ dy ]
        #     [ rz ]    [ -G     H^-1 ]  [ dz ]
        #               \____ = M ____/
        # for
        #     1) (cx, cy, cz) := M \ (c, b, h)
        #     2) (vx, vy, vz) := M \ (rx, ry, rz + H \ rs)
        # (the first one has been precomputed)
        self.solve_sys_3(self.v_xyz, r._xyz, r.s)

        # Second, backsubstitute to obtain solutions for the full 6x6 system
        #    dtau := (rtau + rkap + c' vx + b' vy + h' vz) / (T + c' cx + b' cy + h' cz)
        #     dx  := vx - dtau cx
        #     dy  := vy - dtau cy
        #     dz  := vz - dtau cz
        #     ds  := -G dx + dtau * h - rz
        #    dkap := rkap - T * dtau
        # where T := mu/tau^2, or T := kap/tau if NT scaling is used

        # taunum := rtau + rkap + c' vx + b' vy + h' vz
        tau_num = r.tau + r.kap + self.cbh.inp(self.v_xyz)
        # tauden := kap / tau + c' cx + b' cy + h' cz
        tau_den = pnt.kap / pnt.tau + self.cbh.inp(self.c_xyz)
        # dtau := taunum / tauden
        d.tau[:] = tau_num / tau_den

        # (dx, dy, dz) := (vx, vy, vz) - dtau * (cx, cy, cz)
        d._xyz.vec[:] = sp.linalg.blas.daxpy(
            self.c_xyz.vec, self.v_xyz.vec, a=-d.tau[0, 0]
        )

        # ds := -G dx + dtau * h - rz
        np.multiply(model.h, d.tau[0, 0], out=d.s.vec)
        d.s.vec -= model.G @ d.x
        d.s.vec -= r.z.vec

        # dkap := rkap - (kap/tau) * dtau
        d.kap[:] = r.kap - (pnt.kap / pnt.tau) * d.tau

        return d

    def solve_sys_3(self, d, r, rs=None):
        # Compute (dx, dy, dz) by solving the 3x3 block system
        #     [ rx ]   [    ]    [      A'  G' ]  [ dx ]
        #     [ ry ] + [    ] := [ -A          ]  [ dy ]
        #     [ rz ]   [H\rs]    [ -G     H^-1 ]  [ dz ]
        # where H = mu H(s), or H = H(w) if NT scaling is used,
        # for a given (rx, ry, rz, rs).

        model = self.model

        if model.use_A and model.use_G:
            # In the general case
            #     dy := (A (G'HG)^-1 A') \ [ry + A (G'HG) \ [rx - G' (H rz + rs)]]
            #     dx := (G'HG) \ [rx - G' (H rz + rs) - A' dy]
            #     dz := H (rz + G dx) + rs

            # dy := (A (G'HG)^-1 A') \ [ry + A (G'HG) \ [rx - G' (H rz + rs)]]
            blk_hess_prod_ip(self.work1, r.z, model)
            if rs is not None:
                self.work1 += rs
            temp = r.x - model.G_T @ self.work1.vec
            if self.GHG_issingular:
                temp -= model.A_T @ r.y
            temp = r.y + model.A @ cho_solve(self.GHG_fact, temp)
            d.y[:] = cho_solve(self.AGHGA_fact, temp)

            # dx := (G'HG) \ [rx - G' (H rz + rs) - A' dy]
            temp = r.x - model.G_T @ self.work1.vec - model.A_T @ d.y
            if self.GHG_issingular:
                temp -= model.A_T @ r.y
            d.x[:] = cho_solve(self.GHG_fact, temp)

            # dz := H (rz + G dx) + rs
            d.z.vec[:] = self.work1.vec
            self.work2.vec[:] = model.G @ d.x
            blk_hess_prod_ip(self.work1, self.work2, model)
            d.z.vec += self.work1.vec

        elif model.use_A and not model.use_G:
            # If G = -I (or some easily invertible square diagonal scaling), then
            #     [ rx ]   [    ]    [      A' -I  ]  [ dx ]
            #     [ ry ] + [    ] := [ -A          ]  [ dy ]
            #     [ rz ]   [H\rs]    [  I     H^-1 ]  [ dz ]
            # dy := (AG^-1 H^-1 (AG^-1)') \ [ry + A (G^-1 (H \ [G^-1 rx - rs] - rz)]
            # dz := G^-1 (rx - A' dy)
            # dx := G^-1 (H \ [G^-1 (rx - A' dy) - rs] - rz)

            # dy := (AG^-1 H^-1 (AG^-1)') \ [ry + A (G^-1 (H \ [G^-1 rx - rs] - rz)]
            np.multiply(r.x, model.G_inv, out=self.work1.vec)
            if rs is not None:
                self.work1.vec -= rs.vec
            blk_invhess_prod_ip(self.work2, self.work1, model)
            self.work2.vec -= r.z.vec
            np.multiply(self.work2.vec, model.G_inv, out=d.x)
            temp = model.A @ d.x
            temp += r.y
            d.y[:] = cho_solve(self.AHA_fact, temp)

            # dz := G^-1 (rx - A' dy)
            A_T_dy = model.A_T @ d.y
            np.subtract(r.x, A_T_dy, out=d.z.vec)
            d.z.vec *= model.G_inv

            # dx := G^-1 (H \ [G^-1 rx - rs] - rz - H \ [G^-1 A' dy])
            np.multiply(A_T_dy, model.G_inv, out=self.work1.vec)
            blk_invhess_prod_ip(self.work2, self.work1, model)
            self.work2.vec *= model.G_inv
            d.x -= self.work2.vec

        elif not model.use_A and model.use_G:
            # If A = [] (i.e, no primal linear constraints), then
            #     [ rx ] + [    ] := [       G' ]  [ dx ]
            #     [ rz ]   [H\rs]    [ -G  H^-1 ]  [ dz ]
            # dx := G'HG \ [rx - G' (H rz + rs)]
            # dz := H (rz + G dx) + rs

            # dx := GHG \ [rx - G' (H rz + rs)]
            blk_hess_prod_ip(self.work1, r.z, model)
            if rs is not None:
                self.work1 += rs
            temp = r.x - model.G_T @ self.work1.vec
            d.x[:] = cho_solve(self.GHG_fact, temp)

            # dz := H (rz + G dx) + rs
            self.work1.copy_from(model.G @ d.x)
            self.work1 += r.z
            blk_hess_prod_ip(d.z, self.work1, model)
            if rs is not None:
                d.z += rs

        elif not model.use_A and not model.use_G:
            # If both A = [] and G = -I, then
            #     [ rx ] + [    ] := [       G  ]  [ dx ]
            #     [ rz ]   [H\rs]    [ -G  H^-1 ]  [ dz ]
            # dz := G \ rx
            # dx := -G \ [rz + H \ (rs - dz)]

            # dz := G \ rx
            np.multiply(r.x, model.G_inv, out=d.z.vec)

            # dx := rz + H \ [rx - dz]
            self.work1.vec[:] = -d.z.vec
            if rs is not None:
                self.work1 += rs
            blk_invhess_prod_ip(self.work2, self.work1, model)
            d.x[:] = self.work2.vec
            d.x += r.z.vec
            d.x *= -model.G_inv

        return d

    def apply_sys_6(self, r, d):
        # Compute (rx, ry, rz, rs, rtau, rkap) as a forwards pass
        # of the 6x6 block system
        #     [ rx ]    [      A'  G'  c ]  [ dx ]   [    ]
        #     [ ry ] := [ -A           b ]  [ dy ] - [    ]
        #     [ rz ]    [ -G           h ]  [ dz ]   [ ds ]
        #     [rtau]    [ -c' -b' -h'    ]  [dtau]   [dkap]
        # and
        #       rs   :=  mu H(s) ds + dz
        #      rkap  :=  (mu / tau^2) dtau + dkap
        # or, if NT scaling is used,
        #       rs   :=  H(w) ds + dz
        #      rkap  :=  (kap / tau) dtau + dkap
        # for a given (dx, dy, dz, ds, dtau, dkap).

        model = self.model
        pnt = self.pnt

        # rx := A' dy + G' dz + c dtau
        np.multiply(model.c, d.tau[0, 0], out=r.x)
        r.x += model.A_T @ d.y
        r.x += model.G_T @ d.z.vec

        # ry := -A dx + b dtau
        np.multiply(model.b, d.tau[0, 0], out=r.y)
        r.y -= model.A @ d.x

        # rz := -G dx + h dtau - ds
        np.multiply(model.h, d.tau[0, 0], out=r.z.vec)
        r.z.vec -= model.G @ d.x
        r.z.vec -= d.s.vec

        # rs := mu H ds + dz
        blk_hess_prod_ip(r.s, d.s, model)
        r.s.vec += d.z.vec

        # rtau := -c' dx - b' dy - h' dz - dkap
        r.tau[:] = (
            -(model.c.T @ d.x) - (model.b.T @ d.y) - (model.h.T @ d.z.vec) - d.kap[0, 0]
        )

        # rkap := (kap / tau) dtau + dkap
        r.kap[:] = (pnt.kap / pnt.tau) * d.tau[0, 0] + d.kap[0, 0]

        return r

    def apply_sys_3(self, r, d):
        # Compute (rx, ry, rz) as a forwards pass
        # of the 3x3 block system
        #     [ rx ]    [      A'  G'  ]  [ dx ]
        #     [ ry ] := [ -A           ]  [ dy ]
        #     [ rz ]    [ -G      H^-1 ]  [ dz ]
        # where H = mu H(s), or H = H(w) if NT scaling is used,
        # for a given (rx, ry, rz).

        model = self.model

        # rx := A' dy + G' dz
        r.x[:] = model.A_T @ d.y
        r.x += model.G_T @ d.z.vec

        # ry := -A dx
        r.y[:] = model.A @ d.x
        r.y *= -1

        # pz := -G dx + H \ dz
        blk_invhess_prod_ip(r.z, d.z, model)
        r.z.vec -= model.G @ d.x

        return r

    def solve_sys_6_ir(self, d, r):
        return solve_sys_ir(d, r, self.apply_sys_6, self.solve_sys_6, self.res_pnt,
                            self.ir_pnt, self.ir_settings)  # fmt: skip

    def solve_sys_3_ir(self, d, r):
        return solve_sys_ir(d, r, self.apply_sys_3, self.solve_sys_3, self.res_xyz,
                            self.ir_xyz, self.ir_settings)  # fmt: skip


def solve_sys_ir(x, b, A, A_inv, res, cor, settings):
    # Perform iterative refinement on a solution x of a linear system
    #     A x = b

    ir_maxiter = settings["maxiter"]
    ir_tol = settings["tol"]
    ir_improv_ratio = settings["improv_ratio"]

    r_norm = b.norm()

    # Compute residuals:
    # res := b - A x
    A(res, x)
    res -= b
    res_norm = res.norm() / (1 + r_norm)

    for i in range(ir_maxiter):
        # If solution is accurate enough, exit iterative refinement
        if res_norm < ir_tol:
            break
        prev_res_norm = res_norm

        # Perform iterative refinement
        # cor := A \ res
        A_inv(cor, res)
        cor *= -1
        # x := x + cor
        cor += x

        # Check residuals
        #     res := b - A x
        A(res, cor)
        res -= b
        res_norm = res.norm() / (1 + r_norm)

        # Exit if iterative refinement made things worse
        improv_ratio = prev_res_norm / (res_norm + 1e-15)
        if improv_ratio < 1:
            return prev_res_norm

        # Otherwise, update solution
        x.vec[:] = cor.vec

        # Exit if iterative refinement is slow
        if improv_ratio < ir_improv_ratio:
            break

    return res_norm


def blk_hess_prod_ip(out, dirs, model):
    if model.issymmetric:
        for k, cone_k in enumerate(model.cones):
            cone_k.nt_prod_ip(out[k], dirs[k])
    else:
        for k, cone_k in enumerate(model.cones):
            cone_k.hess_prod_ip(out[k], dirs[k])
    return out


def blk_invhess_prod_ip(out, dirs, model):
    if model.issymmetric:
        for k, cone_k in enumerate(model.cones):
            cone_k.invnt_prod_ip(out[k], dirs[k])
    else:
        for k, cone_k in enumerate(model.cones):
            cone_k.invhess_prod_ip(out[k], dirs[k])
    return out


def blk_hess_congruence(dirs, model):
    n = model.n
    out = np.zeros((n, n))

    if model.issymmetric:
        for k, cone_k in enumerate(model.cones):
            out += cone_k.nt_congr(dirs[k])
    else:
        for k, cone_k in enumerate(model.cones):
            out += cone_k.hess_congr(dirs[k])

    return out


def blk_invhess_congruence(dirs, model):
    p = model.p
    out = np.zeros((p, p))

    if model.issymmetric:
        for k, cone_k in enumerate(model.cones):
            out += cone_k.invnt_congr(dirs[k])
    else:
        for k, cone_k in enumerate(model.cones):
            out += cone_k.invhess_congr(dirs[k])

    return out



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics

import os

import numpy as np
import scipy as sp

from qics import Model
from qics.cones import (
    ClassEntr,
    ClassRelEntr,
    NonNegOrthant,
    OpPerspecEpi,
    OpPerspecTr,
    PosSemidefinite,
    QuantCondEntr,
    QuantEntr,
    QuantKeyDist,
    QuantRelEntr,
    SecondOrder,
)
from qics.vectorize import mat_dim, vec_to_mat


def read_file(filename):
    """Reads a file representing a conic program, and return a
    :class:`~qics.Model` representing this problem. Currently supports
    ``*.dat-s``, ``*.dat-c``, and ``*.cbf`` file formats.

    Parameters
    ----------
    filename : :obj:`string`
        Name of the file and file format we want to read.

    Returns
    -------
    :class:`~qics.Model`
        Model representing the conic program from the specified file.

    See Also
    --------
    write_file : Write a conic program to a file of a specified format.
    read_sdpa : Read file in the SDPA sparse format.
    read_cbf : Read file in the Conic Benchmark Format.
    """
    file_extension = os.path.splitext(filename)[1]
    if file_extension == ".dat-s" or file_extension == ".dat-c":
        return read_sdpa(filename)
    if file_extension == ".cbf":
        return read_cbf(filename)
    raise Exception("Unsupported file extension.")


def write_file(model, filename):
    """Writes a conic program represented by a :class:`~qics.Model`
    to a specified file and format. Currently supports ``*.dat-s``,
    ``*.dat-c``, and ``*.cbf`` formats.

    Parameters
    ----------
    model : :class:`~qics.Model`
        Model representing the conic program we want to write to a file.
    filename : :obj:`string`
        Name of the file and file format we want to write to.

    See Also
    --------
    read_file : Read conic program from a file of a specified format.
    write_sdpa : Write file in the SDPA sparse format.
    write_cbf : Write file in the Conic Benchmark Format.
    """
    file_extension = os.path.splitext(filename)[1]
    if file_extension == ".dat-s" or file_extension == ".dat-c":
        return write_sdpa(model, filename)
    if file_extension == ".cbf":
        return write_cbf(model, filename)
    raise Exception("Unsupported file extension.")


def read_sdpa(filename):
    r"""Reads a file in the SDPA sparse format, and returns a
    :class:`~qics.Model` represnting the standard form semidefinite program

    .. math::

        \min_{x \in \mathbb{R}^p} &&& c^\top x

        \text{s.t.} &&& \sum_{i=1}^p F_i x_i - F_0 \succeq 0

    Two types of SDPA sparse file formats are supported.

    - ``*.dat-s``: Standard SDPA sparse file, where
      :math:`F_i\in\mathbb{S}^n` for :math:`i=0,\ldots,p`.
    - ``*.dat-c``: Complex-valued SDPA sparse file, where
      :math:`F_i\in\mathbb{H}^n` for :math:`i=0,\ldots,p`.

    Parameters
    ----------
    filename : :obj:`string`
        Name of the SPDA sparse file we want to read.

    Returns
    -------
    :class:`~qics.Model`
        Model representing the semidefinite program from the specified
        file.

    See Also
    --------
    write_sdpa : Write file in the SDPA sparse format.
    """

    # Determine if this is a complex or real SDP file
    file_extension = os.path.splitext(filename)[1]
    assert file_extension == ".dat-s" or file_extension == ".dat-c"
    iscomplex = file_extension[-1] == "c"

    f = open(filename, "r")
    line = f.readline()

    ##############################################################
    # Skip comments
    ##############################################################
    # From user manual:
    #   On top of the input data file, we can write a single or
    #   multiple lines of Title and Comment. Each line of Title
    #   and Comment must begin with " or * and consist of no more
    #   than 75 letters;
    while line[0] == "*" or line[0] == '"':
        line = f.readline()

    ##############################################################
    # Read mDim (number of linear constraints b)
    ##############################################################
    mDim = int(line.strip().split(" ")[0])

    ##############################################################
    # Read nBlock (number of blocks in X and Z)
    ##############################################################
    line = f.readline()
    # nBlock = int(line.strip().split(' ')[0])

    ##############################################################
    # Read blockStruct (structure of blocks in X and Z; negative
    # integer represents a diagonal block)
    ##############################################################
    line = f.readline()
    blockStruct = [int(i) for i in line.strip().split(" ")]

    ##############################################################
    # Read b
    ##############################################################
    line = f.readline()
    line = line.strip()
    line = line.strip("{}()")
    if "," in line:
        b_str = line.strip().split(",")
    else:
        b_str = line.strip().split()
    while b_str.count("") > 0:
        b_str.remove("")
    b = np.array([[float(bi)] for bi in b_str])

    ##############################################################
    # Read c and A
    ##############################################################
    # Some useful dimension information
    step = 2 if iscomplex else 1
    dims = [step * n * n if n >= 0 else -n for n in blockStruct]
    idxs = np.insert(np.cumsum(dims), 0, 0)
    totDim = sum(dims)

    # Preallocate c
    c = np.zeros((totDim, 1))
    C = []
    for i, ni in enumerate(blockStruct):
        if ni >= 0:
            if iscomplex:
                C.append(
                    c[idxs[i] : idxs[i + 1]]
                    .reshape((-1, 2))
                    .view(dtype=np.complex128)
                    .reshape(ni, ni)
                )
            else:
                C.append(c[idxs[i] : idxs[i + 1]].reshape((ni, ni)))
        else:
            # Real vector
            C.append(c[idxs[i] : idxs[i + 1]])

    # Preallocate A (we will build a sparse matrix)
    Acols = []
    Arows = []
    Avals = []

    lineList = f.readlines()
    for line in lineList:
        row, block, colI, colJ, val = line.split()[0:5]
        row = int(row.strip(",")) - 1
        block = int(block.strip(",")) - 1
        colI = int(colI.strip(",")) - 1
        colJ = int(colJ.strip(",")) - 1

        ni = blockStruct[block]
        val = (
            complex(val.strip(","))
            if (iscomplex and ni >= 0)
            else float(val.strip(","))
        )

        if val == 0:
            continue

        if row == -1:
            # First row corresponds to data for c
            if ni >= 0:
                # Symmetric or Hermitian matrix
                C[block][colI, colJ] = val
                C[block][colJ, colI] = np.conj(val)
            else:
                # Real vector
                assert colI == colJ
                C[block][colI] = val
        else:
            # All other rows correspond to data for A
            if ni >= 0:
                # Symmetric or Hermitian matrix
                if val.real != 0.0:
                    Acols.append(idxs[block] + (colI + colJ * ni) * step)
                    Arows.append(row)
                    Avals.append(val.real)

                    if colJ != colI:
                        Acols.append(idxs[block] + (colJ + colI * ni) * step)
                        Arows.append(row)
                        Avals.append(val.real)

                if val.imag != 0.0:
                    # Hermitian matrices should have real diagonal entries
                    assert colI != colJ
                    assert iscomplex
                    Acols.append(idxs[block] + (colI + colJ * ni) * step + 1)
                    Arows.append(row)
                    Avals.append(-val.imag)

                    Acols.append(idxs[block] + (colJ + colI * ni) * step + 1)
                    Arows.append(row)
                    Avals.append(val.imag)
            else:
                # Real vector
                assert colI == colJ
                Acols.append(idxs[block] + colI)
                Arows.append(row)
                Avals.append(val)

    c *= -1  # SDPA format maximizes c
    A = sp.sparse.csr_matrix((Avals, (Arows, Acols)), shape=(mDim, totDim))

    ##############################################################
    # Get cones
    ##############################################################
    cones = []
    for bi in blockStruct:
        if bi >= 0:
            cones.append(PosSemidefinite(bi, iscomplex=iscomplex))
        else:
            cones.append(NonNegOrthant(-bi))

    return Model(c=c, A=A, b=b, cones=cones)


def write_sdpa(model, filename):
    r"""Writes a standard form semidefinite program, i.e.,

    .. math::

        \min_{x \in \mathbb{R}^p} &&& c^\top x

        \text{s.t.} &&& \sum_{i=1}^p F_i x_i - F_0 \succeq 0,

    or

    .. math::

        \min_{X \in \mathbb{S}^n} &&& \langle F_0, X \rangle

        \text{s.t.} &&& \langle F_i, X \rangle = b_i, \quad i=1,\ldots,p

        &&& X \succeq 0,

    represented by a :class:`~qics.Model` to a file in the SDPA sparse
    format ``*.dat-s``. If any of the matrices :math:`F_i` are complex
    for :math:`i=0,\ldots,p`, then we use the complex SDPA sparse format
    ``*.dat-c``.

    Parameters
    ----------
    model : :class:`~qics.Model`
        Model representing the standard form semidefinite program we want
        to write to a file.
    filename : :obj:`string`
        Name of the SDPA sparse file we want to write to.

    See Also
    --------
    write_sdpa : Read file in the SDPA sparse format.
    """

    # Get SDP data from model
    assert (model.use_A) != (model.use_G)
    c = model.c if model.use_A else model.h
    A = model.A if model.use_A else model.G.T
    b = model.b if model.use_A else -model.c
    cones = model.cones
    iscomplex = model.iscomplex
    if not sp.sparse.issparse(A):
        A = sp.sparse.coo_matrix(A)
    assert iscomplex == (filename[-1] == "c")

    f = open(filename, "w")

    # Write mDim (length of A)
    mDim = b.size
    f.write(str(mDim) + "\n")

    # Write nBlock (number of blocks in the X variable)
    nBlock = len(cones)
    f.write(str(nBlock) + "\n")

    # Write blockStruct (structure of blocks in X variable)
    blockStruct = []
    for cone_k in cones:
        if isinstance(cone_k, PosSemidefinite):
            blockStruct.append(cone_k.n)
        elif isinstance(cone_k, NonNegOrthant):
            blockStruct.append(-cone_k.n)
        else:
            raise Exception("Invalid SDP: model contains unsupported cones.")
    f.write(" ".join(str(ni) for ni in blockStruct) + "\n")

    # Write b
    b_str = ""
    for bi in b.ravel():
        b_str += str(bi) + " "
    f.write(b_str + "\n")

    # Some useful dimension information
    dims = [cone_k.dim for cone_k in cones]
    idxs = np.insert(np.cumsum(dims), 0, 0)

    # Write C
    # Write in format (k, blk, i, j, v), where k=0, blk=block, (i, j)=index, v=value
    for blk in range(nBlock):
        # Turn vectorised Cl into matrix (make diagonal if corresponds to LP)
        Cl = c[idxs[blk] : idxs[blk + 1]]
        if isinstance(cones[blk], PosSemidefinite):
            Cl = vec_to_mat(Cl, iscomplex=cones[blk].get_iscomplex(), compact=False)
        elif isinstance(cones[blk], NonNegOrthant):
            Cl = np.diag(Cl.ravel())
        Cl = sp.sparse.coo_matrix(Cl)

        # Write upper triangular component of matrix
        for i, j, v in zip(Cl.row, Cl.col, Cl.data):
            if i <= j:
                v_str = str(-v).replace("(", "").replace(")", "")
                f.write("0 " + str(blk + 1) + " " + str(i + 1) + " " + str(j + 1)
                        + " " + v_str + "\n")  # fmt: skip

    # Write A
    # Write in format (k, blk, i, j, v), where k=0, blk=block, (i, j)=index, v=value
    A = A.tocsr()
    for k in range(mDim):
        for blk in range(nBlock):
            Akl = A[k, idxs[blk] : idxs[blk + 1]]

            if cones[blk].get_iscomplex():
                Akl = Akl[:, ::2] + Akl[:, 1::2] * 1j

            for idx, v in zip(Akl.indices, Akl.data):
                if isinstance(cones[blk], PosSemidefinite):
                    (i, j) = (idx // cones[blk].n, idx % cones[blk].n)
                else:
                    (i, j) = (idx, idx)

                if i <= j:
                    v_str = str(v).replace("(", "").replace(")", "")
                    f.write(str(k + 1) + " " + str(blk + 1) + " " + str(i + 1)
                        + " " + str(j + 1) + " " + v_str + "\n")  # fmt: skip

    f.close()

    return


def read_cbf(filename):
    r"""Reads a file in the Conic Benchmark Format ``*.cbf``, and returns a
    :class:`~qics.Model` representing a conic program of the form

    .. math::

        \min_{x \in \mathbb{R}^n} &&& c^\top x

        \text{s.t.} &&& b - Ax = 0

         &&& h - Gx \in \mathcal{K}.

    .. warning::

        This function is quite limited in the types of ``*.cbf`` files that
        can be read. It is recommended to only use this function together
        with files written using :func:`~qics.io.write_cbf`.

    Parameters
    ----------
    filename : :obj:`string`
        Name of the CBF file we want to read.

    Returns
    -------
    :class:`~qics.Model`
        Model representing the conic program from the specified file.

    See Also
    --------
    write_cbf : Write file in the Conic Benchmark Format.
    """
    # Determine if this is a complex or real SDP file
    file_extension = os.path.splitext(filename)[1]
    assert file_extension == ".cbf"

    f = open(filename, "r")
    cones = []
    offset = 0.0
    lookup = {"qce": [], "qkd": [], "mgm": []}

    def _read_cones(cone_type, cone_dim, lookup):
        if cone_type == "L+":
            return NonNegOrthant(cone_dim)
        elif cone_type == "Q":
            n = 1 - cone_dim
            return SecondOrder(n)
        elif cone_type == "SVECPSD":
            n = mat_dim(cone_dim, compact=True)
            return PosSemidefinite(n)
        elif cone_type == "HVECPSD":
            n = mat_dim(cone_dim, iscomplex=True, compact=True)
            return PosSemidefinite(n, iscomplex=True)
        elif cone_type == "CE":
            n = cone_dim - 2
            return ClassEntr(n)
        elif cone_type == "CRE":
            n = (cone_dim - 1) // 2
            return ClassRelEntr(n)
        elif cone_type == "SVECQE":
            n = mat_dim(cone_dim - 2, compact=True)
            return QuantEntr(n)
        elif cone_type == "HVECQE":
            n = mat_dim(cone_dim - 2, iscomplex=True, compact=True)
            return QuantEntr(n, iscomplex=True)
        elif cone_type == "SVECQRE":
            n = mat_dim((cone_dim - 1) // 2, compact=True)
            return QuantRelEntr(n)
        elif cone_type == "HVECQRE":
            n = mat_dim((cone_dim - 1) // 2, iscomplex=True, compact=True)
            return QuantRelEntr(n, iscomplex=True)
        elif "SVECQCE" in cone_type:
            lookup_id = int(cone_type[1])
            dims = lookup["qce"][lookup_id][0]
            sys = lookup["qce"][lookup_id][1]
            return QuantCondEntr(dims, sys)
        elif "HVECQCE" in cone_type:
            lookup_id = int(cone_type[1])
            dims = lookup["qce"][lookup_id][0]
            sys = lookup["qce"][lookup_id][1]
            return QuantCondEntr(dims, sys, iscomplex=True)
        elif "SVECQKD" in cone_type:
            lookup_id = int(cone_type[1])
            G_info = lookup["qkd"][lookup_id][0]
            Z_info = lookup["qkd"][lookup_id][1]
            return QuantKeyDist(G_info, Z_info)
        elif "HVECQKD" in cone_type:
            lookup_id = int(cone_type[1])
            G_info = lookup["qkd"][lookup_id][0]
            Z_info = lookup["qkd"][lookup_id][1]
            return QuantKeyDist(G_info, Z_info, iscomplex=True)
        elif cone_type == "SVECORE":
            n = mat_dim(cone_dim // 3, compact=True)
            return OpPerspecEpi(n, "log")
        elif cone_type == "HVECORE":
            n = mat_dim(cone_dim // 3, iscomplex=True, compact=True)
            return OpPerspecEpi(n, "log", iscomplex=True)
        elif "SVECMGM" in cone_type:
            lookup_id = int(cone_type[1])
            power = lookup["mgm"][lookup_id]
            n = mat_dim(cone_dim // 3, compact=True)
            return OpPerspecEpi(n, power)
        elif "HVECMGM" in cone_type:
            lookup_id = int(cone_type[1])
            power = lookup["mgm"][lookup_id]
            n = mat_dim(cone_dim // 3, iscomplex=True, compact=True)
            return OpPerspecEpi(n, power, iscomplex=True)
        elif cone_type == "SVECTRE":
            n = mat_dim((cone_dim - 1) // 2, compact=True)
            return OpPerspecTr(n, "log")
        elif cone_type == "HVECTRE":
            n = mat_dim((cone_dim - 1) // 2, iscomplex=True, compact=True)
            return OpPerspecTr(n, "log", iscomplex=True)
        elif "SVECTGM" in cone_type:
            lookup_id = int(cone_type[1])
            power = lookup["mgm"][lookup_id]
            n = mat_dim((cone_dim - 1) // 2, compact=True)
            return OpPerspecTr(n, power)
        elif "HVECTGM" in cone_type:
            lookup_id = int(cone_type[1])
            power = lookup["mgm"][lookup_id]
            n = mat_dim((cone_dim - 1) // 2, iscomplex=True, compact=True)
            return OpPerspecTr(n, power, iscomplex=True)

    while True:
        line = f.readline()
        if not line:
            break
        keyword = line.strip()
        if keyword == "" or keyword[0] == "#":
            continue

        ########################
        ## File information
        ########################
        if keyword == "VER":
            ver = int(f.readline().strip())
            if ver != 4:
                print("Warning: Version of .cbf file not supported.")

        ########################
        ## Model structure
        ########################
        if keyword == "OBJSENSE":
            line = f.readline().strip()
            if line == "MIN":
                objsense = 1
            elif line == "MAX":
                objsense = -1
            else:
                raise Exception(
                    "Invalid OBJSENSE read from .cbf file (must be MIN or MAX)."
                )

        if keyword == "QCECONES":
            line = f.readline().strip()
            (ncones, _) = [int(i) for i in line.strip().split(" ")]
            for i in range(ncones):
                _ = int(f.readline().strip())
                line = f.readline()
                dims_k = [int(i) for i in line.strip().split(" ")]
                sys_k = int(f.readline().strip())
                lookup["qce"] += [(dims_k, sys_k)]

        if keyword == "QKDCONES":
            line = f.readline().strip()
            (ncones, totalparam) = [int(i) for i in line.strip().split(" ")]
            for k in range(ncones):
                line = f.readline().strip()
                (nnz_k, Klen_k, Kdim0_k, Kdim1_k, iscomplex_k) = [
                    int(i) for i in line.strip().split(" ")
                ]
                dtype = np.complex128 if iscomplex_k else np.float64
                K_list = [
                    np.zeros((Kdim0_k, Kdim1_k), dtype=dtype) for _ in range(Klen_k)
                ]
                for _ in range(nnz_k):
                    line = f.readline().strip().split(" ")
                    if iscomplex_k:
                        K_list[int(line[0])][int(line[1]), int(line[2])] = complex(
                            float(line[3]), float(line[4])
                        )
                    else:
                        K_list[int(line[0])][int(line[1]), int(line[2])] = float(
                            line[3]
                        )

                line = f.readline().strip()
                (nnz_k, Zlen_k, Zdim0_k, Zdim1_k, _) = [
                    int(i) for i in line.strip().split(" ")
                ]
                Z_list = [np.zeros((Zdim0_k, Zdim1_k)) for _ in range(Zlen_k)]
                for _ in range(nnz_k):
                    line = f.readline().strip().split(" ")
                    Z_list[int(line[0])][int(line[1]), int(line[2])] = float(line[3])

                lookup["qkd"] += [(K_list, Z_list)]

        if keyword == "MGMCONES":
            line = f.readline().strip()
            (ncones, _) = [int(i) for i in line.strip().split(" ")]
            for i in range(ncones):
                _ = int(f.readline().strip())
                power_k = float(f.readline().strip())
                lookup["mgm"] += [power_k]

        if keyword == "VAR":
            # Number and domain of variables
            # i.e., variables of the form x \in K
            line = f.readline()
            (nx, ncones) = [int(i) for i in line.strip().split(" ")]
            for i in range(ncones):
                line = f.readline().strip().split(" ")
                (cone_type, cone_dim) = (line[0], int(line[1]))
                if cone_type == "F":
                    # Corresponds to use_G = True
                    assert ncones == 1
                    assert nx == cone_dim
                    use_G = True
                else:
                    cones += [_read_cones(cone_type, cone_dim, lookup)]
                    use_G = False

        if keyword == "CON":
            # Number and domain of affine constrained variables
            # i.e., variables of the form Ax-b \in K
            line = f.readline()
            total_cone_dim = 0
            A_idxs = np.arange(0)
            (ng, ncones) = [int(i) for i in line.strip().split(" ")]
            for i in range(ncones):
                line = f.readline().strip().split(" ")
                (cone_type, cone_dim) = (line[0], int(line[1]))
                if cone_type == "L=":
                    A_idxs = np.arange(total_cone_dim, total_cone_dim + cone_dim)
                else:
                    cones += [_read_cones(cone_type, cone_dim, lookup)]
                total_cone_dim += cone_dim

        ########################
        ## Problem data
        ########################
        if keyword == "OBJACOORD":
            # Sparse objective
            c = np.zeros((nx, 1))
            nnz = int(f.readline().strip())
            for i in range(nnz):
                line = f.readline().strip().split(" ")
                c[int(line[0])] = float(line[1])

        if keyword == "OBJBCOORD":
            # Objective offset
            offset = float(f.readline().strip())

        if keyword == "ACOORD":
            # Sparse constraint matrix A
            A = np.zeros((ng, nx))
            nnz = int(f.readline().strip())
            for k in range(nnz):
                line = f.readline().strip().split(" ")
                A[int(line[0]), int(line[1])] = float(line[2])

        if keyword == "BCOORD":
            # Sparse constraint vector b
            b = np.zeros((ng, 1))
            nnz = int(f.readline().strip())
            for i in range(nnz):
                line = f.readline().strip().split(" ")
                b[int(line[0])] = float(line[1])

    ########################
    ## Process problem data
    ########################
    if use_G:
        T = _get_expand_compact_matrices(cones)

        # Need to split A into [-G; -A] and b into [-h; -b]
        # and uncompact G, h
        c *= objsense
        G_idxs = np.delete(np.arange(A.shape[0]), A_idxs)
        G = -T.T @ A[G_idxs]
        h = -T.T @ b[G_idxs]
        A = A[A_idxs]
        b = b[A_idxs]
    else:
        T = _get_expand_compact_matrices(cones)

        # No G, just need to uncompact c and A
        c = T.T @ c * objsense
        A = A @ T
        G = None
        h = None

    return Model(c=c, A=A, b=b, G=G, h=h, cones=cones, offset=offset)


def write_cbf(model, filename):
    r"""Writes a conic program

    .. math::

        \min_{x \in \mathbb{R}^n} &&& c^\top x

        \text{s.t.} &&& b - Ax = 0

         &&& h - Gx \in \mathcal{K}

    represented by a :class:`~qics.Model` to a ``.cbf`` file using the
    Conic Benchmark Format.

    Parameters
    ----------
    model : :class:`~qics.Model`
        Model representing the conic program we want to write to a CBF
        file.
    filename : :obj:`string`
        Name of the CBF file we want to write to.

    See Also
    --------
    write_cbf : Read file in the Conic Benchmark Format.
    """
    if model.use_G:
        T = _get_expand_compact_matrices(model.cones)

        c = model.c.copy()
        A = model.A.copy()
        b = model.b.copy()
        G = T @ model.G
        h = T @ model.h
    else:
        T = _get_expand_compact_matrices(model.cones)

        c = T @ model.c
        A = model.A @ T.T
        b = model.b.copy()
    cones = model.cones

    f = open(filename, "w")

    ########################
    ## File information
    ########################
    # Version number
    f.write("VER" + "\n")
    f.write(str(4) + "\n\n")

    ########################
    ## Model structure
    ########################
    # Objective sense
    f.write("OBJSENSE" + "\n")
    f.write("MIN" + "\n\n")

    # Constraints
    q = 0
    cones_string = ""
    lookup = {"qce": [], "qkd": [], "mgm": []}
    for cone_k in cones:
        if isinstance(cone_k, NonNegOrthant):
            (cone_name, cone_size) = ("L+", cone_k.n)
        if isinstance(cone_k, SecondOrder):
            (cone_name, cone_size) = ("Q", 1 + cone_k.n)
        if isinstance(cone_k, PosSemidefinite):
            if cone_k.get_iscomplex():
                (cone_name, cone_size) = ("HVECPSD", cone_k.n * cone_k.n)
            else:
                (cone_name, cone_size) = ("SVECPSD", cone_k.n * (cone_k.n + 1) // 2)
        if isinstance(cone_k, ClassEntr):
            (cone_name, cone_size) = ("CE", 2 + cone_k.n)
        if isinstance(cone_k, ClassRelEntr):
            (cone_name, cone_size) = ("CRE", 1 + 2 * cone_k.n)
        if isinstance(cone_k, QuantEntr):
            if cone_k.get_iscomplex():
                (cone_name, cone_size) = ("HVECQE", 2 + cone_k.n * cone_k.n)
            else:
                (cone_name, cone_size) = ("SVECQE", 2 + cone_k.n * (cone_k.n + 1) // 2)
        if isinstance(cone_k, QuantRelEntr):
            if cone_k.get_iscomplex():
                (cone_name, cone_size) = ("HVECQRE", 1 + 2 * cone_k.n * cone_k.n)
            else:
                (cone_name, cone_size) = ("SVECQRE", 1 + cone_k.n * (cone_k.n + 1))
        if isinstance(cone_k, QuantCondEntr):
            if cone_k.get_iscomplex():
                cone_name = "@" + str(len(lookup["qce"])) + ":HVECQCE"
                cone_size = 1 + cone_k.N * cone_k.N
            else:
                cone_name = "@" + str(len(lookup["qce"])) + ":SVECQCE"
                cone_size = 1 + cone_k.N * (cone_k.N + 1) // 2
            lookup["qce"] += [(cone_k.dims, cone_k.sys)]
        if isinstance(cone_k, QuantKeyDist):
            if cone_k.get_iscomplex():
                cone_name = "@" + str(len(lookup["qkd"])) + ":HVECQKD"
                cone_size = 1 + cone_k.n * cone_k.n
            else:
                cone_name = "@" + str(len(lookup["qkd"])) + ":SVECQKD"
                cone_size = 1 + cone_k.n * (cone_k.n + 1) // 2
            lookup["qkd"] += [(cone_k.K_list_raw, cone_k.Z_list_raw)]
        if isinstance(cone_k, OpPerspecTr):
            if cone_k.get_iscomplex():
                if cone_k.func == "log":
                    cone_name = "HVECTRE"
                else:
                    cone_name = "@" + str(len(lookup["mgm"])) + ":HVECTGM"
                    lookup["mgm"] += [cone_k.func]
                cone_size = 1 + 2 * cone_k.n * cone_k.n
            else:
                if cone_k.func == "log":
                    cone_name = "SVECTRE"
                else:
                    cone_name = "@" + str(len(lookup["mgm"])) + ":SVECTGM"
                    lookup["mgm"] += [cone_k.func]
                cone_size = 1 + 2 * cone_k.n * (cone_k.n + 1) // 2
        if isinstance(cone_k, OpPerspecEpi):
            if cone_k.get_iscomplex():
                if cone_k.func == "log":
                    cone_name = "HVECORE"
                else:
                    cone_name = "@" + str(len(lookup["mgm"])) + ":HVECMGM"
                    lookup["mgm"] += [cone_k.func]
                cone_size = 3 * cone_k.n * cone_k.n
            else:
                if cone_k.func == "log":
                    cone_name = "SVECORE"
                else:
                    cone_name = "@" + str(len(lookup["mgm"])) + ":SVECMGM"
                    lookup["mgm"] += [cone_k.func]
                cone_size = 3 * cone_k.n * (cone_k.n + 1) // 2

        q += cone_size
        cones_string += cone_name + " " + str(cone_size) + "\n"

    # Lookup table
    if len(lookup["qce"]) > 0:
        f.write("QCECONES" + "\n")
        f.write(str(len(lookup["qce"])) + " " + str(2 * len(lookup["qce"])) + "\n")
        for k, qce_k in enumerate(lookup["qce"]):
            f.write(str(2) + "\n")
            f.write(" ".join(str(ni) for ni in qce_k[0]) + "\n")
            f.write(str(qce_k[1]) + "\n")
        f.write("\n")

    if len(lookup["qkd"]) > 0:
        f.write("QKDCONES" + "\n")
        total_nnz = sum(
            [
                sum([np.count_nonzero(Ki) for Ki in qkd_k[0]])
                + sum([np.count_nonzero(Zi) for Zi in qkd_k[1]])
                for qkd_k in lookup["qkd"]
            ]
        )
        f.write(str(len(lookup["qkd"])) + " " + str(total_nnz) + "\n")
        for k, qkd_k in enumerate(lookup["qkd"]):
            # Total nnz, how many Klist, and size of Klist, and if complex or not
            totalnnz_k = sum([np.count_nonzero(Ki) for Ki in qkd_k[0]])
            iscomplex_k = any([np.iscomplexobj(Ki) for Ki in qkd_k[0]])
            dtype = np.complex128 if iscomplex_k else np.float64
            f.write(str(totalnnz_k) + " " + str(len(qkd_k[0])) + " "
                + str(qkd_k[0][0].shape[0]) + " " + str(qkd_k[0][0].shape[1])
                + " " + str(int(iscomplex_k)) + "\n")  # fmt: skip
            for t, Kt in enumerate(qkd_k[0]):
                # Write Ki
                Kt = sp.sparse.coo_matrix(Kt, dtype=dtype)
                for it, jt, vt in zip(Kt.row, Kt.col, Kt.data):
                    if iscomplex_k:
                        f.write(str(t) + " " + str(it) + " " + str(jt) + " "
                            + str(vt.real) + " " + str(vt.imag) + "\n")  # fmt: skip
                    else:
                        f.write(str(t) + " " + str(it) + " " + str(jt) + " "
                            + str(vt) + "\n")  # fmt: skip

            # How many Zlist, and size of Zlist, and if complex or not
            totalnnz_k = sum([np.count_nonzero(Zi) for Zi in qkd_k[1]])
            f.write(str(totalnnz_k) + " " + str(len(qkd_k[1])) + " "
                + str(qkd_k[1][0].shape[0]) + " " + str(qkd_k[1][0].shape[1]) + " 0\n")  # fmt: skip
            for t, Zt in enumerate(qkd_k[1]):
                # Write Zi
                Zt = sp.sparse.coo_matrix(Zt, dtype=np.float64)
                for it, jt, vt in zip(Zt.row, Zt.col, Zt.data):
                    f.write(
                        str(t) + " " + str(it) + " " + str(jt) + " " + str(vt) + "\n"
                    )
        f.write("\n")

    if len(lookup["mgm"]) > 0:
        f.write("MGMCONES" + "\n")
        f.write(str(len(lookup["mgm"])) + " " + str(len(lookup["mgm"])) + "\n")
        for mgm_k in lookup["mgm"]:
            f.write(str(1) + "\n")
            f.write(str(mgm_k) + "\n")
        f.write("\n")

    if model.use_G:
        f.write("VAR" + "\n")
        f.write(str(model.n) + " " + str(1) + "\n")
        f.write("F" + " " + str(model.n) + "\n\n")

        f.write("CON" + "\n")
        f.write(str(q + model.p) + " " + str(len(cones) + model.use_A) + "\n")
    else:
        f.write("VAR" + "\n")
        f.write(str(q) + " " + str(len(cones)) + "\n")

    f.write(cones_string)

    if model.use_A:
        if not model.use_G:
            f.write("\n" + "CON" + "\n")
            f.write(str(model.p) + " " + str(1) + "\n")

        f.write("L=" + " " + str(model.p) + "\n")

    ########################
    ## Problem data
    ########################
    c = sp.sparse.csc_matrix(c)
    f.write("\n" + "OBJACOORD" + "\n")
    f.write(str(c.nnz) + "\n")
    for ik, vk in zip(c.indices, c.data):
        f.write(str(ik) + " " + str(vk) + "\n")

    if model.offset != 0.0:
        f.write("\n" + "OBJBCOORD" + "\n")
        f.write(str(model.offset) + "\n")

    if model.use_G:
        if not sp.sparse.issparse(G):
            G = sp.sparse.coo_matrix(G)
        if not sp.sparse.issparse(A):
            A = sp.sparse.coo_matrix(A)
        A = sp.sparse.coo_matrix(sp.sparse.vstack([-G, A]))
    else:
        A = sp.sparse.coo_matrix(A)
    f.write("\n" + "ACOORD" + "\n")
    f.write(str(A.nnz) + "\n")
    for ik, jk, vk in zip(A.row, A.col, A.data):
        f.write(str(ik) + " " + str(jk) + " " + str(vk) + "\n")

    if model.use_G:
        b = sp.sparse.csc_matrix(np.vstack((-h, b)))
    else:
        b = sp.sparse.csc_matrix(b)
    f.write("\n" + "BCOORD" + "\n")
    f.write(str(b.nnz) + "\n")
    for ik, vk in zip(b.indices, b.data):
        f.write(str(ik) + " " + str(vk) + "\n")

    f.close()

    return


def _get_compact_to_full_op(n, iscomplex=False):
    import scipy

    dim_compact = n * n if iscomplex else n * (n + 1) // 2
    dim_full = 2 * n * n if iscomplex else n * n

    rows = np.zeros(dim_full)
    cols = np.zeros(dim_full)
    vals = np.zeros(dim_full)

    irt2 = np.sqrt(0.5)

    row = 0
    k = 0
    for j in range(n):
        for i in range(j):
            rows[k : k + 2] = row
            cols[k : k + 2] = (
                [2 * (i + j * n), 2 * (j + i * n)]
                if iscomplex
                else [i + j * n, j + i * n]
            )
            vals[k : k + 2] = irt2
            k += 2
            row += 1

            if iscomplex:
                rows[k : k + 2] = row
                cols[k : k + 2] = [2 * (i + j * n) + 1, 2 * (j + i * n) + 1]
                vals[k : k + 2] = [-irt2, irt2]
                k += 2
                row += 1

        rows[k] = row
        cols[k] = 2 * j * (n + 1) if iscomplex else j * (n + 1)
        vals[k] = 1.0
        k += 1
        row += 1

    return scipy.sparse.csr_matrix((vals, (rows, cols)), shape=(dim_compact, dim_full))


def _get_expand_compact_matrices(cones):
    # Split A into columns correpsonding to each variable
    compact_to_full_op = []
    for cone_k in cones:
        for type_k, dim_k in zip(cone_k.type, cone_k.dim):
            if type_k == "s":
                n = int(np.sqrt(dim_k))
                compact_to_full_op += [_get_compact_to_full_op(n)]
            elif type_k == "h":
                n = int(np.sqrt(dim_k // 2))
                compact_to_full_op += [_get_compact_to_full_op(n, True)]
            else:
                compact_to_full_op += [sp.sparse.eye(dim_k)]

    return sp.sparse.block_diag(compact_to_full_op, format="csc")



# Copyright (c) 2024, Kerry He, James Saunderson, and Hamza Fawzi

# This Python package QICS is licensed under the MIT license; see LICENSE.md
# file in the root directory or at https://github.com/kerry-he/qics
